# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime  # noqa: F401
from decimal import Decimal
from typing import Any, Dict, List, Optional, Sequence, Union, cast  # noqa: F401

from PIL import Image

from correios.exceptions import (
    InvalidExtraServiceError,
    InvalidFederalTaxNumberError,
    InvalidRegionalDirectionError,
    InvalidServiceDeclaredValue,
    InvalidUserContractError,
    MaximumDeclaredValueError,
    MinimumDeclaredValueError,
)

from ..utils import get_resource_path, to_datetime, to_integer
from .data import (
    EXTRA_SERVICE_VD_PAC,
    EXTRA_SERVICE_VD_PAC_MINI,
    EXTRA_SERVICE_VD_SEDEX,
    EXTRA_SERVICES,
    REGIONAL_DIRECTIONS,
    SERVICES,
)

EXTRA_SERVICE_CODE_SIZE = 2


def _to_federal_tax_number(federal_tax_number) -> "FederalTaxNumber":
    if isinstance(federal_tax_number, FederalTaxNumber):
        return federal_tax_number

    return FederalTaxNumber(federal_tax_number)


def _to_state_tax_number(state_tax_number) -> "StateTaxNumber":
    if isinstance(state_tax_number, StateTaxNumber):
        return state_tax_number

    return StateTaxNumber(state_tax_number)


class AbstractTaxNumber:
    def __init__(self, number: str) -> None:
        self._number = self._validate(number)

    @property
    def number(self) -> str:
        return self._number

    def _sanitize(self, raw_number: str) -> str:
        return "".join(d for d in raw_number if d.isdigit())

    def _validate(self, raw_number: str):
        raise NotImplementedError()  # pragma: no cover

    def display(self) -> str:
        raise NotImplementedError()  # pragma: no cover

    def __str__(self):
        return self.number

    def __repr__(self):
        return "<{} number: {}>".format(self.__class__.__name__, self.number)

    def __eq__(self, other):
        return self.number == self._sanitize(other)


class FederalTaxNumber(AbstractTaxNumber):
    FEDERAL_TAX_NUMBER_SIZE = 14

    def _check_verification_digits(self, raw_number):
        number = [int(d) for d in raw_number[:12]]
        prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
        assert len(number) == len(prod)

        while len(number) < 14:
            r = sum(x * y for (x, y) in zip(number, prod)) % 11
            f = (11 - r) if r > 1 else 0
            number.append(f)
            prod.insert(0, prod[0] + 1)

        number = "".join(str(d) for d in number)
        return raw_number == number

    def _validate(self, raw_number: str):
        raw_number = self._sanitize(raw_number)

        if len(raw_number) != FederalTaxNumber.FEDERAL_TAX_NUMBER_SIZE:
            raise InvalidFederalTaxNumberError("Tax Number must have {} digits".format(self.FEDERAL_TAX_NUMBER_SIZE))

        if not self._check_verification_digits(raw_number):
            raise InvalidFederalTaxNumberError("Invalid Federal Tax Number verification digits")

        return raw_number

    def display(self) -> str:
        return "{}.{}.{}/{}-{}".format(
            self.number[:2], self.number[2:5], self.number[5:8], self.number[8:12], self.number[12:]
        )


class StateTaxNumber(AbstractTaxNumber):
    def _validate(self, raw_number: str):
        return self._sanitize(raw_number)

    def display(self) -> str:
        return self.number


class Service:
    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        code: Union[int, str],
        id: int,
        description: str,
        category: str,
        display_name: Optional[str] = "",
        symbol: Optional[str] = None,
        max_weight: Optional[int] = None,
        min_declared_value: Decimal = Decimal("0.00"),
        max_declared_value: Decimal = Decimal("0.00"),
        default_extra_services: Optional[Sequence[Union["ExtraService", int]]] = None,
    ) -> None:
        self.id = id
        self.code = Service.sanitize_code(code)
        self.description = description.strip()
        self.display_name = display_name or self.description
        self.category = category.strip()
        self.symbol = symbol or "economic"
        self._symbol_image = None  # type: Optional[Image]
        self.max_weight = max_weight
        self.min_declared_value = min_declared_value
        self.max_declared_value = max_declared_value

        if default_extra_services is None:
            self.default_extra_services = []  # type: List
        else:
            self.default_extra_services = [ExtraService.get(es) for es in default_extra_services]

    def __str__(self):
        return str(self.code)

    def __repr__(self):
        return "<Service code={!r}, name={!r}>".format(self.code, self.display_name)

    def __eq__(self, other):
        other = Service.get(other)
        return (self.id, self.code) == (other.id, other.code)

    def validate_insurance_declared_value(self, value: Union[Decimal, float], insurance_code: int) -> bool:
        services_need_value_gt_zero = [EXTRA_SERVICE_VD_SEDEX, EXTRA_SERVICE_VD_PAC, EXTRA_SERVICE_VD_PAC_MINI]
        if value == 0 and insurance_code in services_need_value_gt_zero:
            msg = "Value cannot be zero for this insurance code : {}".format(insurance_code)
            raise InvalidServiceDeclaredValue(msg)
        if insurance_code not in services_need_value_gt_zero and value > 0:
            msg = "Value cannot be greater than zero for this insurance code : {}".format(insurance_code)
            raise InvalidServiceDeclaredValue(msg)
        return True

    def validate_declared_value(self, value: Union[Decimal, float]) -> bool:
        if value > self.max_declared_value:
            raise MaximumDeclaredValueError(
                "Declared value {!r} is greater than maximum "
                "{!r} for service {!r}".format(value, self.max_declared_value, self)
            )
        if value < self.min_declared_value:
            raise MinimumDeclaredValueError(
                "Declared value {!r} is less than minimum "
                "{!r} for service {!r}".format(value, self.min_declared_value, self)
            )
        return True

    def get_symbol_filename(self, extension="gif") -> str:
        filename = "{}.{}".format(self.symbol, extension)
        return str(get_resource_path(filename))

    @property
    def symbol_image(self) -> Image.Image:
        if not self._symbol_image:
            self._symbol_image = Image.open(self.get_symbol_filename())
        return self._symbol_image

    @classmethod
    def sanitize_code(cls, code: Union[int, str]) -> str:
        code = to_integer("".join(d for d in str(code) if d.isdigit()))
        return "{:05}".format(code)

    @classmethod
    def get(cls, service: Union["Service", int, str]) -> "Service":
        if isinstance(service, cls):
            return service
        service = cast(Union[int, str], service)
        code = cls.sanitize_code(service)
        return cls(code=code, **SERVICES[code])


class ExtraService:
    def __init__(self, number: int, code: str, name: str, display_on_label: bool = True) -> None:
        if not number:
            raise InvalidExtraServiceError("Invalid Extra Service Number {!r}".format(number))
        self.number = number

        if not code or len(code) != EXTRA_SERVICE_CODE_SIZE:
            raise InvalidExtraServiceError("Invalid Extra Service Code {!r}".format(code))
        self.code = code.upper()

        if not name:
            raise InvalidExtraServiceError("Invalid Extra Service Name {!r}".format(name))
        self.name = name

        self.display_on_label = display_on_label

    def __repr__(self):
        return "<ExtraService number={!r}, code={!r}>".format(self.number, self.code)

    def __eq__(self, other):
        if isinstance(other, int):
            return self.number == other
        return self.number == other.number

    def is_declared_value(self):
        return self in (EXTRA_SERVICE_VD_PAC, EXTRA_SERVICE_VD_PAC_MINI, EXTRA_SERVICE_VD_SEDEX)

    @classmethod
    def get(cls, number: Union["ExtraService", int]) -> "ExtraService":
        if isinstance(number, cls):
            return number
        number = cast(int, number)
        attrs = EXTRA_SERVICES[number]  # type: Dict[str, Any]
        return cls(number=number, **attrs)


class User:
    def __init__(
        self,
        name: str,
        federal_tax_number: Union[str, FederalTaxNumber],
        state_tax_number: Optional[Union[str, StateTaxNumber]] = None,
        status_number: Optional[Union[int, str]] = None,
    ) -> None:
        self.name = name.strip()
        self.federal_tax_number = _to_federal_tax_number(federal_tax_number)

        if status_number is not None:
            status_number = to_integer(status_number)
        self.status_number = status_number

        if state_tax_number is not None:
            state_tax_number = _to_state_tax_number(state_tax_number)
        self.state_tax_number = state_tax_number

        self.contracts = []  # type: List[Contract]

    def add_contract(self, contract: "Contract"):
        if contract in self.contracts:
            raise InvalidUserContractError("Contract {!r} already added".format(contract))
        self.contracts.append(contract)


class Contract:
    def __init__(
        self, user: User, number: Union[int, str], regional_direction: Union[str, int, "RegionalDirection"]
    ) -> None:

        self.user = user
        user.add_contract(self)

        self.number = to_integer(number)

        if isinstance(regional_direction, str):
            regional_direction = int(regional_direction)

        if isinstance(regional_direction, int):
            regional_direction = RegionalDirection.get(regional_direction)

        self.regional_direction = regional_direction

        self._customer_code = None  # type: Optional[str]
        self._start_date = None  # type: Optional[datetime]
        self._end_date = None  # type: Optional[datetime]
        self.status_code = None  # type: Optional[str]
        self.posting_cards = []  # type: List[PostingCard]

    @property
    def customer_code(self):
        return self._customer_code

    @customer_code.setter
    def customer_code(self, code):
        self._customer_code = to_integer(code)

    @property
    def start_date(self):
        return self._start_date

    @start_date.setter
    def start_date(self, date):
        self._start_date = to_datetime(date)

    @property
    def end_date(self):
        return self._end_date

    @end_date.setter
    def end_date(self, date):
        self._end_date = to_datetime(date)

    def add_posting_card(self, posting_card: "PostingCard"):
        posting_card.contract = self
        self.posting_cards.append(posting_card)

    @property
    def regional_direction_number(self):
        return self.regional_direction.number

    @property
    def customer_name(self) -> str:
        return self.user.name

    def __str__(self):
        return str(self.number)

    def __repr__(self):
        return "<Contract number={!r}>".format(self.number)


class PostingCard:
    ACTIVE = True
    CANCELLED = False

    def __init__(
        self, contract: Contract, number: Union[int, str], administrative_code: Union[int, str]  # 10 digits
    ) -> None:  # 8 digits
        self.contract = contract
        self._number = to_integer(number)
        self._administrative_code = to_integer(administrative_code)
        self._start_date = None  # type: Optional[datetime]
        self._end_date = None  # type: Optional[datetime]
        self._status = None  # type: Optional[int]
        self._unit = None  # type: Optional[str]
        self.status_code = None  # type: Optional[str]
        self.services = []  # type: List[Service]

        if self not in self.contract.posting_cards:
            self.contract.add_posting_card(self)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, number):
        self._status = to_integer(number)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, number):
        self._unit = to_integer(number)

    @property
    def number(self):
        return "{:010}".format(self._number)

    @property
    def administrative_code(self):
        return "{:08}".format(self._administrative_code)

    @property
    def start_date(self):
        return self._start_date

    @start_date.setter
    def start_date(self, date):
        self._start_date = to_datetime(date)

    @property
    def end_date(self):
        return self._end_date

    @end_date.setter
    def end_date(self, date):
        self._end_date = to_datetime(date)

    def add_service(self, service: Service):
        self.services.append(service)

    def get_contract_number(self):
        return self.contract.number

    def __repr__(self):
        return "<PostingCard number={!r}, contract={!r}>".format(self.number, self.get_contract_number())

    def __str__(self):
        return self.number


class RegionalDirection:
    def __init__(self, number: int, code: str, name: str) -> None:
        if not number:
            raise InvalidRegionalDirectionError("Invalid regional direction number {!r}".format(number))

        if not code:
            raise InvalidRegionalDirectionError("Invalid regional direction code {!r}".format(code))

        if not name:
            raise InvalidRegionalDirectionError("Invalid regional direction name {!r}".format(name))

        self.number = to_integer(number)
        self.code = code.upper()
        self.name = name

    @classmethod
    def get(cls, number: Union["RegionalDirection", int]) -> "RegionalDirection":
        if isinstance(number, RegionalDirection):
            return number

        return cls(number=number, **REGIONAL_DIRECTIONS[number])

    def __repr__(self):
        return "<RegionalDirection number={!r}, code={!r}>".format(self.number, self.code)



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings
from decimal import Decimal

import pytest

from correios.exceptions import InvalidStateError, InvalidZipCodeError
from correios.models.address import Address, Phone, ReceiverAddress, SenderAddress, State, ZipCode


def test_basic_zip():
    zip_code = ZipCode("82940150")
    assert zip_code.code == "82940150"
    assert zip_code.prefix == 82940
    assert zip_code.sufix == 150


def test_sanitize_zip():
    zip_code = ZipCode("82940-150")
    assert zip_code.code == "82940150"
    assert zip_code.prefix == 82940
    assert zip_code.sufix == 150


@pytest.mark.parametrize("zipcode", ("12345", "123456789", "12.345-000", "12345.000"))
def test_fail_invalid_zipcode_format(zipcode):
    with pytest.raises(InvalidZipCodeError):
        ZipCode(zipcode)


@pytest.mark.parametrize("zipcode", ("00000-000", "728001-000"))
def test_fail_invalid_zipcode_range(zipcode):
    with pytest.raises(InvalidZipCodeError):
        ZipCode(zipcode)


@pytest.mark.parametrize(
    "zip_code, state",
    [
        ("13560060", "SP"),
        ("04547-003", "SP"),
        ("28921-100", "RJ"),
        ("22440-033", "RJ"),
        ("29213-360", "ES"),
        ("29048-495", "ES"),
        ("35900-457", "MG"),
        ("30190-060", "MG"),
        ("42800-040", "BA"),
        ("40015-170", "BA"),
        ("49500-000", "SE"),
        ("49037-580", "SE"),
        ("57820-000", "AL"),
        ("57032-901", "AL"),
        ("53370-255", "PE"),
        ("50030-150", "PE"),
        ("58400-355", "PB"),
        ("58051-900", "PB"),
        ("59580-000", "RN"),
        ("59015-900", "RN"),
        ("64260-000", "PI"),
        ("64046-902", "PI"),
        ("62598-973", "CE"),
        ("60060-440", "CE"),
        ("65268-000", "MA"),
        ("65056-480", "MA"),
        ("68040-090", "PA"),
        ("66115-970", "PA"),
        ("68925-000", "AP"),
        ("68908-119", "AP"),
        ("69340-000", "RR"),
        ("69310-000", "RR"),
        ("69400-000", "AM"),
        ("69041-000", "AM"),
        ("69928-000", "AC"),
        ("69918-093", "AC"),
        ("76907-438", "RO"),
        ("76803-970", "RO"),
        ("78903-972", "RO"),
        ("77600-000", "TO"),
        ("77061-900", "TO"),
        ("76400-000", "GO"),
        ("74810-907", "GO"),
        ("73010-521", "DF"),
        ("70150-000", "DF"),
        ("78200-000", "MT"),
        ("78030-210", "MT"),
        ("79730-000", "MS"),
        ("79101-901", "MS"),
        ("83203-100", "PR"),
        ("80730-000", "PR"),
        ("88330-000", "SC"),
        ("88058-512", "SC"),
        ("95670-000", "RS"),
        ("90560-003", "RS"),
    ],
)
def test_zip_code_state(zip_code, state):
    zip_code = ZipCode(zip_code)
    assert zip_code.state == state


@pytest.mark.parametrize(
    "zip_code, region",
    [
        ("13560060", "INTERIOR"),
        ("04547-003", "CAPITAL"),
        ("28921-100", "INTERIOR"),
        ("22440-033", "CAPITAL"),
        ("29213-360", "INTERIOR"),
        ("29048-495", "CAPITAL"),
        ("35900-457", "INTERIOR"),
        ("30190-060", "CAPITAL"),
        ("42800-040", "INTERIOR"),
        ("40015-170", "CAPITAL"),
        ("49500-000", "INTERIOR"),
        ("49037-580", "CAPITAL"),
        ("57820-000", "INTERIOR"),
        ("57032-901", "CAPITAL"),
        ("53370-255", "INTERIOR"),
        ("50030-150", "CAPITAL"),
        ("58400-355", "INTERIOR"),
        ("58051-900", "CAPITAL"),
        ("59580-000", "INTERIOR"),
        ("59015-900", "CAPITAL"),
        ("64260-000", "INTERIOR"),
        ("64046-902", "CAPITAL"),
        ("62598-973", "INTERIOR"),
        ("60060-440", "CAPITAL"),
        ("65268-000", "INTERIOR"),
        ("65056-480", "CAPITAL"),
        ("68040-090", "INTERIOR"),
        ("66115-970", "CAPITAL"),
        ("68925-000", "INTERIOR"),
        ("68908-119", "CAPITAL"),
        ("69340-000", "INTERIOR"),
        ("69310-000", "CAPITAL"),
        ("69400-000", "INTERIOR"),
        ("69041-000", "CAPITAL"),
        ("69928-000", "INTERIOR"),
        ("69918-093", "CAPITAL"),
        ("76907-438", "INTERIOR"),
        ("76803-970", "CAPITAL"),
        ("78903-972", "CAPITAL"),
        ("77600-000", "INTERIOR"),
        ("77061-900", "CAPITAL"),
        ("76400-000", "INTERIOR"),
        ("74810-907", "CAPITAL"),
        ("73010-521", "CAPITAL"),
        ("70150-000", "CAPITAL"),
        ("78200-000", "INTERIOR"),
        ("78030-210", "CAPITAL"),
        ("79730-000", "INTERIOR"),
        ("79101-901", "CAPITAL"),
        ("83203-100", "INTERIOR"),
        ("80730-000", "CAPITAL"),
        ("88330-000", "INTERIOR"),
        ("88058-512", "CAPITAL"),
        ("95670-000", "INTERIOR"),
        ("90560-003", "CAPITAL"),
    ],
)
def test_zip_code_region(zip_code, region):
    zip_code = ZipCode(zip_code)
    assert zip_code.region == region


def test_convert_zip_to_str():
    assert str(ZipCode("82940-150")) == "82940150"


def test_zip_repr():
    assert repr(ZipCode("82940-150")) == "<ZipCode code: 82940150>"


def test_zip_display():
    assert ZipCode("82940150").display() == "82940-150"


@pytest.mark.parametrize("zip_code,digit", [("71010050", 6), ("82940150", 1)])
def test_zip_code_check_digit(zip_code, digit):
    zip_code = ZipCode(zip_code)
    assert zip_code.digit == digit


@pytest.mark.parametrize("zip_code", [71010050, "71010050", ZipCode("71010050")])
def test_zip_code_creation(zip_code):
    assert ZipCode.create(zip_code) == ZipCode("71010050")


def test_basic_phone():
    phone = Phone("+1 (212) 555-1234")
    assert phone == "+12125551234"
    assert phone == Phone("12125551234", "US")
    assert phone.display() == "212-555-1234"
    assert phone.short == "2125551234"
    assert str(phone) == "12125551234"
    assert repr(phone) == "<Phone 12125551234>"


def test_empty_phone():
    phone = Phone("")
    assert phone == ""
    assert phone == Phone("", "US")
    assert phone.display() == ""
    assert phone.short == ""
    assert str(phone) == ""
    assert repr(phone) == "<Phone >"


@pytest.mark.parametrize(
    "code,name",
    (
        ("AC", "Acre"),
        ("AL", "Alagoas"),
        ("AP", "Amapá"),
        ("AM", "Amazonas"),
        ("BA", "Bahia"),
        ("CE", "Ceará"),
        ("DF", "Distrito Federal"),
        ("ES", "Espírito Santo"),
        ("GO", "Goiás"),
        ("MA", "Maranhão"),
        ("MT", "Mato Grosso"),
        ("MS", "Mato Grosso do Sul"),
        ("MG", "Minas Gerais"),
        ("PA", "Pará"),
        ("PB", "Paraíba"),
        ("PR", "Paraná"),
        ("PE", "Pernambuco"),
        ("PI", "Piauí"),
        ("RJ", "Rio de Janeiro"),
        ("RN", "Rio Grande do Norte"),
        ("RS", "Rio Grande do Sul"),
        ("RO", "Rondônia"),
        ("RR", "Roraima"),
        ("SC", "Santa Catarina"),
        ("SP", "São Paulo"),
        ("SE", "Sergipe"),
        ("TO", "Tocantins"),
    ),
)
def test_states(code, name):
    state = State(code)
    assert state.code == code
    assert state.display() == name

    state = State(name)
    assert state.code == code
    assert state.display() == name


@pytest.mark.parametrize("state", ("df", "distrito federal"))
def test_lowercase_state_name(state):
    state = State(state)
    assert state.code == "DF"
    assert state.display() == "Distrito Federal"


def test_fail_invalid_state():
    with pytest.raises(InvalidStateError):
        State("XY")

    with pytest.raises(InvalidStateError):
        State("Unknown State")


def test_convert_state():
    assert str(State("Distrito Federal")) == "DF"


def test_state_repr():
    assert repr(State("DF")) == "<State code: DF name: Distrito Federal>"


def test_basic_address():
    address = Address(
        name="John Doe",
        phone="+111555-12345",
        cellphone="+11155-54321",
        email="john.doe@example.com",
        street="Rua dos Bobos",
        number="5",
        complement="apto. 3",
        neighborhood="Centro",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        latitude="10.0",
        longitude="-10.0",
    )

    assert address.name == "John Doe"
    assert address.phone == "+111555-12345"
    assert address.cellphone == "+11155-54321"
    assert address.email == "john.doe@example.com"
    assert address.street == "Rua dos Bobos"
    assert address.number == "5"
    assert address.complement == "apto. 3"
    assert address.neighborhood == "Centro"
    assert address.city == "Vinicius de Moraes"
    assert address.state == "RJ"
    assert address.zip_code == "12345-678"
    assert address.zip_code_display == "12345-678"
    assert address.basic_address == "Rua Dos Bobos, 5 - Apto. 3, Centro"


def test_basic_address_only_mandatory_args():
    address = Address(
        name="JOHN DOE", street="Rua dos Bobos", number="0", city="Vinicius de Moraes", state="RJ", zip_code="12345-678"
    )

    assert address.name == "JOHN DOE"
    assert address.street == "Rua dos Bobos"
    assert address.number == "0"
    assert address.city == "Vinicius de Moraes"
    assert address.state == "RJ"
    assert address.zip_code == "12345-678"
    assert address.email == ""
    assert address.complement == ""
    assert address.neighborhood == ""
    assert address.phone == ""
    assert address.cellphone == ""
    assert address.latitude == Decimal("0.0")
    assert address.longitude == Decimal("0.0")
    assert address.basic_address == "Rua Dos Bobos, 0"
    assert address.label_name == "John Doe"


def test_basic_address_with_neighborhood():
    address = Address(
        name="John Doe",
        street="Rua dos Bobos",
        number="5",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        neighborhood="Vila Nau",
        complement="ap 5",
    )

    assert address.basic_address == "Rua Dos Bobos, 5 - Ap 5, Vila Nau"


def test_basic_address_with_neighborhood_without_complement():
    address = Address(
        name="John Doe",
        street="Rua dos Bobos",
        number="5",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        neighborhood="Vila Nau",
    )

    assert address.basic_address == "Rua Dos Bobos, 5, Vila Nau"


@pytest.mark.parametrize(
    "raw,filtered,number,zip_complement",
    (
        ("1234", "1234", "1234", "1234"),
        ("123B", "123", "123", "123"),
        ("km 5", "5", "5", "5"),
        ("s/n", "", "S/N", "0"),
        ("S/N", "", "S/N", "0"),
    ),
)
def test_address_number_handling(raw, filtered, number, zip_complement):
    address = Address(
        name="John Doe", street="Rua dos Bobos", city="Vinicius de Moraes", state="RJ", zip_code="12345-678", number=raw
    )
    assert address.filtered_number == filtered
    assert address.number == number
    assert address.zip_complement == zip_complement


def test_address_label_address():
    address = Address(
        name="John Doe",
        street="RUA dos Bobos",
        number="1234",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        neighborhood="VILA Vileza",
        complement="AP 01",
    )

    with warnings.catch_warnings(record=True) as captured_warnings:
        warnings.simplefilter("always")

        assert "Rua" in address.label_address
        assert "Vila" in address.label_address
        assert "1234" in address.label_address
        assert "Ap 01" in address.label_address

        assert len(captured_warnings) == 4
        assert all(w.category == DeprecationWarning for w in captured_warnings)
        assert all("deprecated" in str(w.message) for w in captured_warnings)


@pytest.mark.parametrize("address_class", (ReceiverAddress, SenderAddress))
def test_custom_address_label_address(address_class):
    address = address_class(
        name="John Doe",
        street="RUA dos Bobos",
        number="1234",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        neighborhood="VILA Vileza",
        complement="AP 01",
    )

    assert "<br/>" in address.label_address
    assert "Rua" in address.label_address
    assert "Vila" in address.label_address
    assert "1234" in address.label_address
    assert "Ap 01" in address.label_address


@pytest.mark.parametrize("address_class", (ReceiverAddress, SenderAddress))
def test_custom_address_label_address_long_street_name(address_class):
    address = address_class(
        name="John Doe",
        street="RUA Professor José Caetano dos Santos Mascarenhas",
        number="1234",
        city="Vinicius de Moraes",
        state="RJ",
        zip_code="12345-678",
        neighborhood="VILA Vileza",
        complement="AP 01",
    )

    assert "<br/>" not in address.label_address
    assert "Rua" in address.label_address
    assert "Vila" in address.label_address
    assert "1234" in address.label_address
    assert "Ap 01" in address.label_address



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from typing import Dict, Tuple  # noqa

from ..utils import RangeSet

TRACKING_PREFIX = {
    "AL": "Agentes de leitura",
    "AR": "Avisos de recebimento",
    "AS": "PAC - Ação Social",
    "CA": "Encomenda Internacional - Colis",
    "CB": "Encomenda Internacional - Colis",
    "CC": "Encomenda Internacional - Colis",
    "CD": "Encomenda Internacional - Colis",
    "CE": "Encomenda Internacional - Colis",
    "CF": "Encomenda Internacional - Colis",
    "CG": "Encomenda Internacional - Colis",
    "CH": "Encomenda Internacional - Colis",
    "CI": "Encomenda Internacional - Colis",
    "CJ": "Encomenda Internacional - Colis",
    "CK": "Encomenda Internacional - Colis",
    "CL": "Encomenda Internacional - Colis",
    "CM": "Encomenda Internacional - Colis",
    "CN": "Encomenda Internacional - Colis",
    "CO": "Encomenda Internacional - Colis",
    "CP": "Encomenda Internacional - Colis",
    "CQ": "Encomenda Internacional - Colis",
    "CR": "Carta registrada sem Valor Declarado",
    "CS": "Encomenda Internacional - Colis",
    "CT": "Encomenda Internacional - Colis",
    "CU": "Encomenda internacional - Colis",
    "CV": "Encomenda Internacional - Colis",
    "CW": "Encomenda Internacional - Colis",
    "CX": "Encomenda internacional - Colis ou Selo Lacre para Caixetas",
    "CY": "Encomenda Internacional - Colis",
    "CZ": "Encomenda Internacional - Colis",
    "DA": "SEDEX ou Remessa Expressa com AR Digital",
    "DB": "SEDEX ou Remessa Expressa com AR Digital (Bradesco)",
    "DC": "Remessa Expressa CRLV/CRV/CNH e Notificações",
    "DD": "Devolução de documentos",
    "DE": "Remessa Expressa Talão/Cartão com AR",
    "DF": "e-SEDEX",
    "DG": "SEDEX",
    "DI": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DJ": "SEDEX",
    "DK": "PAC Extra Grande",
    "DL": "SEDEX",
    "DM": "e-SEDEX",
    "DN": "SEDEX",
    "DO": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DP": "SEDEX Pagamento na Entrega",
    "DQ": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DR": "Remessa Expressa com AR Digital (Santander)",
    "DS": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DT": "Remessa econômica com AR Digital (DETRAN)",
    "DU": "e-SEDEX",
    "DV": "SEDEX c/ AR digital",
    "DW": "Encomenda SEDEX (Etiqueta Lógica)",
    "DX": "SEDEX 10",
    "EA": "Encomenda Internacional - EMS",
    "EB": "Encomenda Internacional - EMS",
    "EC": "PAC",
    "ED": "Packet Express",
    "EE": "Encomenda Internacional - EMS",
    "EF": "Encomenda Internacional - EMS",
    "EG": "Encomenda Internacional - EMS",
    "EH": "Encomenda Internacional - EMS ou Encomenda com AR Digital",
    "EI": "Encomenda Internacional - EMS",
    "EJ": "Encomenda Internacional - EMS",
    "EK": "Encomenda Internacional - EMS",
    "EL": "Encomenda Internacional - EMS",
    "EM": "Encomenda Internacional - SEDEX Mundi",  # BR Suffix
    # "EM": "Encomenda Internacional - EMS Importação",
    "EN": "Encomenda Internacional - EMS",
    "EO": "Encomenda Internacional - EMS",
    "EP": "Encomenda Internacional - EMS",
    "EQ": "Encomenda de serviço não expressa (ECT)",
    "ER": "Objeto registrado",
    "ES": "e-SEDEX ou EMS",
    "ET": "Encomenda Internacional - EMS",
    "EU": "Encomenda Internacional - EMS",
    "EV": "Encomenda Internacional - EMS",
    "EW": "Encomenda Internacional - EMS",
    "EX": "Encomenda Internacional - EMS",
    "EY": "Encomenda Internacional - EMS",
    "EZ": "Encomenda Internacional - EMS",
    "FA": "FAC registrado",
    "FE": "Encomenda FNDE",
    "FF": "Objeto registrado (DETRAN)",
    "FH": "FAC registrado com AR Digital",
    "FM": "FAC monitorado",
    "FR": "FAC registrado",
    "IA": "Logística Integrada (agendado / avulso)",
    "IC": "Logística Integrada (a cobrar)",
    "ID": "Logística Integrada (devolução de documento)",
    "IE": "Logística Integrada (Especial)",
    "IF": "CPF",
    "II": "Logística Integrada (ECT)",
    "IK": "Logística Integrada com Coleta Simultânea",
    "IM": "Logística Integrada (Medicamentos)",
    "IN": "Correspondência e EMS recebido do Exterior",
    "IP": "Logística Integrada (Programada)",
    "IR": "Impresso Registrado",
    "IS": "Logística integrada standard (medicamentos)",
    "IT": "Remessa Expressa Medicamentos / Logística Integrada Termolábil",
    "IU": "Logística Integrada (urgente)",
    "IX": "EDEI Expresso",
    "JA": "Remessa econômica com AR Digital",
    "JB": "Remessa econômica com AR Digital",
    "JC": "Remessa econômica com AR Digital",
    "JD": "Remessa econômica Talão/Cartão",
    "JE": "Remessa econômica com AR Digital",
    "JF": "Remessa econômica com AR Digital",
    "JG": "Objeto registrado urgente/prioritário",
    "JH": "Objeto registrado urgente / prioritário",
    "JI": "Remessa econômica Talão/Cartão",
    "JJ": "Objeto registrado (Justiça)",
    "JK": "Remessa econômica Talão/Cartão",
    "JL": "Objeto registrado",
    "JM": "Mala Direta Postal Especial",
    "JN": "Objeto registrado econômico",
    "JO": "Objeto registrado urgente",
    "JP": "Receita Federal",
    "JQ": "Remessa econômica com AR Digital",
    "JR": "Objeto registrado urgente / prioritário",
    "JS": "Objeto registrado",
    "JT": "Objeto Registrado Urgente",
    "JV": "Remessa Econômica (c/ AR DIGITAL)",
    "LA": "SEDEX com Logística Reversa Simultânea em Agência",
    "LB": "e-SEDEX com Logística Reversa Simultânea em Agência",
    "LC": "Objeto Internacional (Prime)",
    "LE": "Logística Reversa Econômica",
    "LF": "Objeto Internacional (Prime)",
    "LI": "Objeto Internacional (Prime)",
    "LJ": "Objeto Internacional (Prime)",
    "LK": "Objeto Internacional (Prime)",
    "LM": "Objeto Internacional (Prime)",
    "LN": "Objeto Internacional (Prime)",
    "LP": "PAC com Logística Reversa Simultânea em Agência",
    "LS": "SEDEX Logística Reversa",
    "LV": "Logística Reversa Expressa",
    "LX": "Packet Standard / Econômica",
    "LZ": "Objeto Internacional (Prime)",
    "MA": "Serviços adicionais do Telegrama",
    "MB": "Telegrama (balcão)",
    "MC": "Telegrama (Fonado)",
    "MD": "SEDEX Mundi (Documento interno)",
    "ME": "Telegrama",
    "MF": "Telegrama (Fonado)",
    "MK": "Telegrama (corporativo)",
    "ML": "Fecha Malas (Rabicho)",
    "MM": "Telegrama (Grandes clientes)",
    "MP": "Telegrama (Pré-pago)",
    "MR": "AR digital",
    "MS": "Encomenda Saúde",
    "MT": "Telegrama (Telemail)",
    "MY": "Telegrama internacional (entrante)",
    "MZ": "Telegrama (Correios Online)",
    "NE": "Tele Sena resgatada",
    "NX": "EDEI Econômico (não urgente)",
    "PA": "Passaporte",
    "PB": "PAC",
    "PC": "PAC a Cobrar",
    "PD": "PAC",
    "PE": "PAC",
    "PF": "Passaporte",
    "PG": "PAC",
    "PH": "PAC",
    "PI": "PAC",
    "PJ": "PAC",
    "PK": "PAC Extra Grande",
    "PL": "PAC",
    "PN": "PAC Normal",
    "PR": "Reembolso Postal",
    "QQ": "Objeto de teste (SIGEP Web)",
    "RA": "Objeto registrado / prioritário",
    "RB": "Carta registrada",
    "RC": "Carta registrada com Valor Declarado",
    "RD": "Remessa econômica ou objeto registrado (DETRAN)",
    "RE": "Objeto registrado econômico",
    "RF": "Receita Federal",
    "RG": "Objeto registrado",
    "RH": "Objeto registrado com AR Digital",
    "RI": "Objeto registrado internacional prioritário",
    "RJ": "Objeto registrado",
    "RK": "Objeto registrado",
    "RL": "Objeto registrado",
    "RM": "Objeto registrado urgente",
    "RN": "Objeto registrado (SIGEPWEB ou Agência)",
    "RO": "Objeto registrado",
    "RP": "Reembolso Postal",
    "RQ": "Objeto registrado",
    "RR": "Objeto registrado",
    "RS": "Objeto registrado",
    "RT": "Remessa econômica Talão/Cartão",
    "RU": "Objeto registrado (ECT)",
    "RV": "Remessa econômica CRLV/CRV/CNH e Notificações com AR Digital",
    "RW": "Objeto internacional",
    "RX": "Objeto internacional",
    "RY": "Remessa econômica Talão/Cartão com AR Digital",
    "RZ": "Objeto registrado",
    "SA": "SEDEX",
    "SB": "SEDEX 10",
    "SC": "SEDEX a cobrar",
    "SD": "SEDEX ou Remessa Expressa (DETRAN)",
    "SE": "SEDEX",
    "SF": "SEDEX",
    "SG": "SEDEX",
    "SH": "SEDEX com AR Digital / SEDEX ou AR Digital",
    "SI": "SEDEX",
    "SJ": "SEDEX Hoje",
    "SK": "SEDEX",
    "SL": "SEDEX",
    "SM": "SEDEX 12",
    "SN": "SEDEX",
    "SO": "SEDEX",
    "SP": "SEDEX Pré-franqueado",
    "SQ": "SEDEX",
    "SR": "SEDEX",
    "SS": "SEDEX",
    "ST": "Remessa Expressa Talão/Cartão",
    "SU": "Encomenda de serviço expressa (ECT)",
    "SV": "Remessa Expressa CRLV/CRV/CNH e Notificações com AR Digital",
    "SW": "e-SEDEX",
    "SX": "SEDEX 10",
    "SY": "Remessa Expressa Talão/Cartão com AR Digital",
    "SZ": "SEDEX",
    "TC": "Objeto para treinamento",
    "TE": "Objeto para treinamento",
    "TS": "Objeto para treinamento",
    "VA": "Encomendas com valor declarado",
    "VC": "Encomendas",
    "VD": "Encomendas com valor declarado",
    "VE": "Encomendas",
    "VF": "Encomendas com valor declarado",
    "VV": "Objeto internacional",
    "XA": "Aviso de chegada (internacional)",
    "XM": "SEDEX Mundi",
    "XR": "Encomenda SUR Postal Expresso",
    "XX": "Encomenda SUR Postal 24 horas",
}

EXTRA_SERVICES = {
    1: {"code": "AR", "name": "Aviso de Recebimento", "display_on_label": True},
    2: {"code": "MP", "name": "Mão Própria Nacional", "display_on_label": True},
    19: {"code": "VD", "name": "Valor Declarado (Encomendas)", "display_on_label": True},  # Sedex
    25: {"code": "RR", "name": "Registro Nacional", "display_on_label": False},
    64: {"code": "VD", "name": "Valor Declarado (Encomendas)", "display_on_label": True},  # PAC
    65: {"code": "VD", "name": "Valor Declarado (Encomendas)", "display_on_label": True},  # PAC MINI
}

EXTRA_SERVICE_AR = 1
EXTRA_SERVICE_MP = 2
EXTRA_SERVICE_VD_SEDEX = 19
EXTRA_SERVICE_RR = 25
EXTRA_SERVICE_VD_PAC = 64
EXTRA_SERVICE_VD_PAC_MINI = 65
EXTRA_SERVICE_VD = EXTRA_SERVICE_VD_SEDEX  # For backward compatibility


SERVICES = {
    "40215": {
        "id": 104707,
        "description": "SEDEX 10",
        "category": "SERVICO_COM_RESTRICAO",
        "max_weight": 10000,
        "display_name": "SEDEX 10",
        "symbol": "premium",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "81019": {
        "id": 104672,
        "description": "E-SEDEX STANDARD",
        "category": "SERVICO_COM_RESTRICAO",
        "max_weight": 15000,
        "display_name": "E-SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "41068": {
        "id": 109819,
        "description": "PAC CONTRATO AGENCIA",
        "category": "PAC",
        "display_name": "PAC",
        "max_weight": 30000,
        "symbol": "standard",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("3000.00"),
    },
    "04669": {
        "id": 124884,
        "description": "PAC",
        "category": "PAC",
        "display_name": "PAC",
        "max_weight": 30000,
        "symbol": "standard",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("3000.00"),
    },
    "40444": {
        "id": 109811,
        "description": "SEDEX - CONTRATO",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40436": {
        "id": 109810,
        "description": "SEDEX - CONTRATO",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40096": {
        "id": 104625,
        "description": "SEDEX (CONTRATO)",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "04162": {
        "id": 124849,
        "description": "SEDEX CONTRATO AGENCIA",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40380": {
        "id": 109806,
        "description": "SEDEX REVERSO 40096",
        "category": "REVERSO",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40010": {
        "id": 104295,
        "description": "SEDEX A VISTA",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "41211": {
        "id": 113546,
        "description": "PAC - CONTRATO",
        "category": "PAC",
        "display_name": "PAC",
        "max_weight": 30000,
        "symbol": "standard",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("3000.00"),
    },
    "40630": {
        "id": 114976,
        "description": "SEDEX PAGAMENTO NA ENTREGA -",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "04316": {
        "id": 124900,
        "description": "SEDEX CONTRATO - UO",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40916": {
        "id": 118568,
        "description": "SEDEX AGRUPADO II",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40908": {
        "id": 118567,
        "description": "SEDEX AGRUPADO I",
        "category": "SEDEX",
        "max_weight": 30000,
        "display_name": "SEDEX",
        "symbol": "express",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "41300": {
        "id": 120366,
        "description": "PAC GRANDES FORMATOS",
        "category": "SERVICO_COM_RESTRICAO",
        "max_weight": 50000,
        "display_name": "PAC",
        "symbol": "standard",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("3000.00"),
    },
    "04812": {
        "id": 124899,
        "description": "PAC CONTRATO - UO",
        "category": "PAC",
        "display_name": "PAC",
        "max_weight": 30000,
        "symbol": "standard",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("3000.00"),
    },
    "40169": {
        "id": 115218,
        "description": "SEDEX 12",
        "category": "SERVICO_COM_RESTRICAO",
        "max_weight": 10000,
        "display_name": "SEDEX 12",
        "symbol": "premium",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "40290": {
        "id": 108934,
        "description": "SEDEX HOJE",
        "category": "SERVICO_COM_RESTRICAO",
        "max_weight": 10000,
        "display_name": "SEDEX Hoje",
        "symbol": "premium",
        "default_extra_services": [EXTRA_SERVICE_RR],
        "min_declared_value": Decimal("20.50"),
        "max_declared_value": Decimal("10000.00"),
    },
    "10154": {
        "id": 118424,
        "description": "CARTA COMERCIAL  REGISTRADA",
        "category": "CARTA_REGISTRADA",
        "display_name": "Carta Registrada",
    },
    "41246": {
        "id": 115487,
        "description": "REM. CAMPANHA PAPAI NOEL DOS",
        "category": "SEM_CATEGORIA",
        "display_name": "Papai Noel dos Correios",
    },
    "40150": {
        "id": 115136,
        "description": "SERVICO DE PROTOCOLO POSTAL -",
        "category": "SEDEX",
        "display_name": "Protocolo",
    },
    "10065": {
        "id": 109480,
        "description": "CARTA COMERCIAL A FATURAR",
        "category": "CARTA_REGISTRADA",
        "display_name": "Carta Comercial",
    },
}  # type: Dict[str, dict]

SERVICE_PAC = "04669"
SERVICE_PAC_INDUSTRIAL = "04812"
SERVICE_SEDEX = "04162"
SERVICE_SEDEX_INDUSTRIAL = "04316"
SERVICE_SEDEX10 = "40215"
SERVICE_SEDEX12 = "40169"
SERVICE_E_SEDEX = "81019"

INSURANCE_VALUE_THRESHOLD_PAC = Decimal("20.50")  # R$
INSURANCE_VALUE_THRESHOLD_SEDEX = Decimal("20.50")  # R$
INSURANCE_PERCENTUAL_COST = Decimal("0.01")  # 1%

REGIONAL_DIRECTIONS = {
    1: {"code": "AC", "name": "AC - ADMINISTRAÇAO CENTRAL"},
    3: {"code": "ACR", "name": "DR - ACRE"},
    4: {"code": "AL", "name": "DR - ALAGOAS"},
    6: {"code": "AM", "name": "DR - AMAZONAS"},
    5: {"code": "AP", "name": "DR - AMAPÁ"},
    8: {"code": "BA", "name": "DR - BAHIA"},
    10: {"code": "BSB", "name": "DR - BRASÍLIA"},
    12: {"code": "CE", "name": "DR - CEARÁ"},
    14: {"code": "ES", "name": "DR - ESPIRITO SANTO"},
    16: {"code": "GO", "name": "DR - GOIÁS"},
    18: {"code": "MA", "name": "DR - MARANHÃO"},
    20: {"code": "MG", "name": "DR - MINAS GERAIS"},
    22: {"code": "MS", "name": "DR - MATO GROSSO DO SUL"},
    24: {"code": "MT", "name": "DR - MATO GROSSO"},
    28: {"code": "PA", "name": "DR - PARÁ"},
    30: {"code": "PB", "name": "DR - PARAÍBA"},
    32: {"code": "PE", "name": "DR - PERNAMBUCO"},
    34: {"code": "PI", "name": "DR - PIAUÍ"},
    36: {"code": "PR", "name": "DR - PARANÁ"},
    50: {"code": "RJ", "name": "DR - RIO DE JANEIRO"},
    60: {"code": "RN", "name": "DR - RIO GRANDE DO NORTE"},
    26: {"code": "RO", "name": "DR - RONDONIA"},
    65: {"code": "RR", "name": "DR - RORAIMA"},
    64: {"code": "RS", "name": "DR - RIO GRANDE DO SUL"},
    68: {"code": "SC", "name": "DR - SANTA CATARINA"},
    70: {"code": "SE", "name": "DR - SERGIPE"},
    74: {"code": "SPI", "name": "DR - SÃO PAULO INTERIOR"},
    72: {"code": "SPM", "name": "DR - SÃO PAULO"},
    75: {"code": "TO", "name": "DR - TOCANTINS"},
}

TRACKING_EVENT_TYPES = {
    "ERROR": "Evento de erro",  # custom event type for "Not Found" error
    "BDE": "Baixa de distribuição externa",
    "BDI": "Baixa de distribuição interna",
    "BDR": "Baixa corretiva",
    "BLQ": "Bloqueio de objetos",
    "CAR": "Conferência de lista de registro",
    "CD": "Conferência de nota de despacho",
    "CMT": "Chegada de um meio de transporte",
    "CO": "Coleta de objetos",
    "CUN": "Conferência de lista de registro",
    "DO": "Expedição de nota de despacho",
    "EST": "Estorno",
    "FC": "Função complementar",
    "IDC": "Indenização de objetos",
    "LDI": "Lista de distribuição interna",
    "LDE": "Lista de distribuição externa",
    "OEC": "Lista de Objetos Entregues ao Carteiro",
    "PAR": "Conferência Unidade Internacional",
    "PMT": "Partida Meio de Transporte",
    "PO": "Postagem (exceção)",
    "RO": "Expedição de Lista de Registro",
    "TRI": "Triagem",
    "CMR": "Conferência de lista de registro",
}

TRACKING_STATUS = {
    ("ERROR", 0): ("error", "Objeto não encontrado", "", ""),  # custom status for "Not Found" error
    ("BLQ", 1): (
        "shipped",
        "Entrega de objeto bloqueada a pedido do remetente",
        "Objeto em análise de destinação",
        "Acompanhar",
    ),
    ("BLQ", 2): ("shipped", "Tentativa de suspensão da entrega", "Objeto em análise de destinação", "Acompanhar"),
    ("BLQ", 11): (
        "shipped",
        "Solicitação de suspensão de entrega recebida",
        "Solicitação realizada pelo contratante/remetente",
        "Acompanhar",
    ),
    ("BLQ", 21): ("shipped", "Solicitação de suspensão de entrega recebida", "", "Acompanhar"),
    ("BDE", 0): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDI", 0): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDR", 0): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("CAR", 1): ("shipped", "Conferido", "Recebido na unidade de destino", "Acompanhar"),
    ("CD", 0): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("CMT", 0): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("CUN", 0): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("DO", 0): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("LDE", 0): ("shipped", "Objeto saiu para entrega ao remetente", "", "Acompanhar"),
    ("LDI", 0): (
        "waiting_retrieval",
        "Objeto aguardando retirada no endereço indicado",
        "Endereço:",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("OEC", 0): (
        "shipped",
        "Objeto saiu para entrega ao destinatário",
        "",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("PO", 0): ("shipped", "Objeto postado", "", "Acompanhar"),
    ("RO", 0): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("TRI", 0): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("BDE", 1): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDE", 70): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDI", 1): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDR", 1): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("CD", 1): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("CO", 1): ("shipped", "Objeto coletado", "", "Acompanhar"),
    ("CUN", 1): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("DO", 1): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("EST", 1): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 1): (
        "shipped",
        "Objeto será devolvido por solicitação do remetente",
        "",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("FC", 10): (
        "shipped",
        "Objeto recebido na unidade de distribuição",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("FC", 47): (
        "shipped",
        "Objeto será devolvido por solicitação do contratante/remetente",
        "Em tratamento, aguarde.",
        "Acompanhar",
    ),
    ("FC", 29): ("shipped", "Objeto encaminhado para apreensão", "Em tratamento, aguarde.", "Acompanhar",),
    ("IDC", 1): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("LDI", 1): (
        "waiting_retrieval",
        "Objeto aguardando retirada no endereço indicado",
        "Endereço:",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("OEC", 1): ("shipped", "Objeto saiu para entrega ao destinatário", "", "Acompanhar"),
    ("PMT", 1): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("PO", 1): ("shipped", "Objeto postado", "", "Acompanhar"),
    ("RO", 1): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("BDE", 2): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Aguarde: Objeto estará disponível para retirada na unidade a ser informada.",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("BDI", 2): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Aguarde: Objeto estará disponível para retirada na unidade a ser informada.",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("BDR", 2): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Aguarde: Objeto estará disponível para retirada na unidade a ser informada.",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("CD", 2): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("DO", 2): ("shipped", "Objeto encaminhado para", "<nome da cidade>", "Acompanhar"),
    ("EST", 2): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 2): ("shipped", "Objeto com data de entrega agendada", "", "Acompanhar"),
    ("IDC", 2): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("LDI", 2): (
        "waiting_retrieval",
        "Objeto disponível para retirada em Caixa Postal",
        "",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("BDE", 3): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.",
    ),
    ("BDI", 3): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.",
    ),
    ("BDR", 3): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.",
    ),
    ("CD", 3): ("shipped", "Objeto recebido na Unidade dos Correios", "", "Acompanhar"),
    ("EST", 3): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 3): ("shipped", "Objeto mal encaminhado", "Encaminhamento a ser corrigido.", "Acompanhar"),
    ("IDC", 3): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("LDI", 3): (
        "waiting_retrieval",
        "Objeto aguardando retirada no endereço indicado",
        "Endereço:",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("BDE", 4): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Cliente recusou-se a receber",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 4): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Cliente recusou-se a receber",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 4): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Cliente recusou-se a receber",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("EST", 4): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 4): (
        "shipped",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("IDC", 4): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("BDE", 5): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 5): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 5): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("EST", 5): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 5): ("shipped", "Objeto devolvido aos Correios", "", "Acompanhar"),
    ("IDC", 5): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("BDE", 6): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Cliente desconhecido no local",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 6): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Cliente desconhecido no local",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 6): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Cliente desconhecido no local",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("EST", 6): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("IDC", 6): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("BDE", 7): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("BDI", 7): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("BDR", 7): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("FC", 7): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Empresa sem expediente",
        "A entrega deverá ocorrer no próximo dia útil",
        "Acompanhar",
    ),
    ("IDC", 7): ("lost", "Objeto não localizado", "Houve indenização dos valores correspondentes", "Acompanhar"),
    ("BDE", 8): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 8): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 8): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("FC", 8): (
        "shipped",
        "Área com distribuição sujeita a prazo diferenciado",
        "Restrição de entrega domiciliar temporária",
        "Acompanhar",
    ),
    ("BDE", 9): (
        "lost",
        "Objeto não localizado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 9): (
        "lost",
        "Objeto não localizado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 9): (
        "lost",
        "Objeto não localizado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("EST", 9): ("ignore_last_entry", "Favor desconsiderar a informação anterior", "", "Acompanhar"),
    ("FC", 9): ("delivery_unsuccessful", "Remetente não retirou objeto na Unidade dos Correios", "", "Acompanhar"),
    ("LDE", 9): ("shipped", "Objeto saiu para entrega ao remetente", "", "Acompanhar"),
    ("OEC", 9): ("shipped", "Objeto saiu para entrega ao remetente", "", "Acompanhar"),
    ("PO", 9): (
        "shipped",
        "Objeto postado após o horário limite da agência",
        "Objeto sujeito a encaminhamento no próximo dia útil",
        "Acompanhar",
    ),
    ("BDE", 10): (
        "shipped_wrong_address",
        "A entrega não pode ser efetuada - Cliente mudou-se",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 10): (
        "shipped_wrong_address",
        "A entrega não pode ser efetuada - Cliente mudou-se",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 10): (
        "shipped_wrong_address",
        "A entrega não pode ser efetuada - Cliente mudou-se",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 12): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 12): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 12): (
        "waiting_retrieval",
        "Remetente não retirou objeto na Unidade dos Correios",
        "Objeto em análise de destinação",
        "Acionar atendimento dos Correios.",
    ),
    ("LDI", 4): (
        "waiting_retrieval",
        "Objeto aguardando retirada no endereço indicado",
        "Endereço:",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("LDI", 11): (
        "waiting_retrieval",
        "Objeto encaminhado para retirada no endereço indicado",
        "Para retirá-lo, é preciso informar o código do objeto.",
        "Acompanhar",
    ),
    ("LDI", 14): (
        "waiting_retrieval",
        "Objeto aguardando retirada no endereço indicado",
        "Endereço:",
        "Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.",
    ),
    ("BDI", 14): ("shipped", "Desistência de postagem pelo remetente", "", "Acompanhar"),
    ("BDR", 14): ("shipped", "Desistência de postagem pelo remetente", "", "Acompanhar"),
    ("BDR", 15): (
        "shipped",
        "Recebido na unidade de distribuição",
        "Por determinação judicial o objeto será entregue em até 7 dias",
        "Acompanhar",
    ),
    ("PAR", 15): ("delivered", "Objeto recebido em <destino>", "", "Acompanhar"),
    ("PAR", 16): (
        "customs_control",
        "Objeto recebido no Brasil",
        "Objeto sujeito à fiscalização e atraso na entrega",
        "Acompanhar",
    ),
    ("PAR", 17): ("customs_control", "Objeto liberado pela alfândega", "", "Acompanhar"),
    ("PAR", 18): ("shipped", "Objeto recebido na unidade de exportação", "", "Acompanhar"),
    ("BDE", 18): (
        "shipped",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Será realizada nova tentativa de entrega no sábado",
        "Acompanhar",
    ),
    ("BDR", 18): (
        "shipped",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Será realizada nova tentativa de entrega no sábado",
        "Acompanhar",
    ),
    ("BDE", 19): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 19): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 19): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Endereço incorreto",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 20): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Será realizada nova tentativa de entrega",
        "Acompanhar",
    ),
    ("BDI", 20): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Será realizada nova tentativa de entrega",
        "Acompanhar",
    ),
    ("BDR", 20): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Será realizada nova tentativa de entrega",
        "Acompanhar",
    ),
    ("BDE", 21): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 21): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 21): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Carteiro não atendido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 22): ("shipped", "Objeto devolvido aos Correios", "", "Acompanhar"),
    ("BDI", 22): ("shipped", "Objeto devolvido aos Correios", "", "Acompanhar"),
    ("BDR", 22): ("shipped", "Objeto devolvido aos Correios", "", "Acompanhar"),
    ("BDE", 23): ("returned", "Objeto devolvido ao remetente", "Recebido por:", "Acompanhar"),
    ("BDI", 23): ("returned", "Objeto devolvido ao remetente", "Recebido por:", "Acompanhar"),
    ("BDR", 23): ("returned", "Objeto devolvido ao remetente", "Recebido por:", "Acompanhar"),
    ("BDE", 24): ("waiting_retrieval", "Objeto disponível para retirada em Caixa Postal", "", "Acompanhar"),
    ("BDI", 24): ("waiting_retrieval", "Objeto disponível para retirada em Caixa Postal", "", "Acompanhar"),
    ("BDR", 24): ("waiting_retrieval", "Objeto disponível para retirada em Caixa Postal", "", "Acompanhar"),
    ("BDE", 25): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Empresa sem expediente",
        "A entrega deverá ocorrer no próximo dia útil",
        "Acompanhar",
    ),
    ("BDI", 25): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Empresa sem expediente",
        "A entrega deverá ocorrer no próximo dia útil",
        "Acompanhar",
    ),
    ("BDR", 25): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Empresa sem expediente",
        "A entrega deverá ocorrer no próximo dia útil",
        "Acompanhar",
    ),
    ("BDE", 26): (
        "waiting_retrieval",
        "Destinatário não retirou objeto na Unidade dos Correios",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 26): (
        "waiting_retrieval",
        "Destinatário não retirou objeto na Unidade dos Correios",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 26): (
        "waiting_retrieval",
        "Destinatário não retirou objeto na Unidade dos Correios",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 28): (
        "damaged",
        "Objeto e/ou conteúdo avariado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 28): (
        "damaged",
        "Objeto e/ou conteúdo avariado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 28): (
        "damaged",
        "Objeto e/ou conteúdo avariado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 30): ("delivery_unsuccessful", "Saída não efetuada", "Em tratamento, aguarde.", "Acompanhar"),
    ("BDI", 30): ("delivery_unsuccessful", "Saída não efetuada", "Em tratamento, aguarde.", "Acompanhar"),
    ("BDR", 30): ("delivery_unsuccessful", "Saída não efetuada", "Em tratamento, aguarde.", "Acompanhar"),
    ("BDE", 32): ("shipped", "Objeto com data de entrega agendada", "", "Acompanhar"),
    ("BDI", 32): ("shipped", "Objeto com data de entrega agendada", "", "Acompanhar"),
    ("BDR", 32): ("shipped", "Objeto com data de entrega agendada", "", "Acompanhar"),
    ("BDE", 33): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Destinatário não apresentou documento exigido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 33): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Destinatário não apresentou documento exigido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 33): (
        "delivery_rejected",
        "A entrega não pode ser efetuada - Destinatário não apresentou documento exigido",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 34): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Logradouro com numeração irregular",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("BDI", 34): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Logradouro com numeração irregular",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("BDR", 34): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada - Logradouro com numeração irregular",
        "Objeto sujeito a atraso na entrega ou a devolução ao remetente",
        "Acompanhar",
    ),
    ("BDE", 35): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Será realizada nova tentativa de coleta ou entrega",
        "Acompanhar",
    ),
    ("BDI", 35): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Será realizada nova tentativa de coleta ou entrega",
        "Acompanhar",
    ),
    ("BDR", 35): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Será realizada nova tentativa de coleta ou entrega",
        "Acompanhar",
    ),
    ("BDE", 36): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 36): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 36): (
        "delivery_unsuccessful",
        "Coleta ou entrega de objeto não efetuada",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 37): (
        "damaged",
        "Objeto e/ou conteúdo avariado por acidente com veículo",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 37): (
        "damaged",
        "Objeto e/ou conteúdo avariado por acidente com veículo",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 37): (
        "damaged",
        "Objeto e/ou conteúdo avariado por acidente com veículo",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 38): (
        "delivery_unsuccessful",
        "Objeto endereçado à empresa falida",
        "Objeto será encaminhado para entrega ao administrador judicial",
        "Acompanhar",
    ),
    ("BDI", 38): (
        "delivery_unsuccessful",
        "Objeto endereçado à empresa falida",
        "Objeto será encaminhado para entrega ao administrador judicial",
        "Acompanhar",
    ),
    ("BDR", 38): (
        "delivery_unsuccessful",
        "Objeto endereçado à empresa falida",
        "Objeto será encaminhado para entrega ao administrador judicial",
        "Acompanhar",
    ),
    ("BDI", 39): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada",
        "Objeto em análise de destinação",
        "Acompanhar",
    ),
    ("BDR", 39): (
        "delivery_unsuccessful",
        "A entrega não pode ser efetuada",
        "Objeto em análise de destinação",
        "Acompanhar",
    ),
    ("BDE", 40): (
        "customs_control",
        "A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 40): (
        "customs_control",
        "A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 40): (
        "customs_control",
        "A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 41): ("unknown", "A entrega do objeto está condicionada à composição do lote", "", "Acompanhar"),
    ("BDI", 41): ("unknown", "A entrega do objeto está condicionada à composição do lote", "", "Acompanhar"),
    ("BDR", 41): ("unknown", "A entrega do objeto está condicionada à composição do lote", "", "Acompanhar"),
    ("BDE", 42): (
        "unknown",
        "Lote de objetos incompleto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 42): (
        "unknown",
        "Lote de objetos incompleto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 42): (
        "unknown",
        "Lote de objetos incompleto",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 43): (
        "customs_control",
        "Objeto apreendido por órgão de fiscalização ou outro órgão anuente",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 43): (
        "customs_control",
        "Objeto apreendido por órgão de fiscalização ou outro órgão anuente",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 43): (
        "customs_control",
        "Objeto apreendido por órgão de fiscalização ou outro órgão anuente",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 45): (
        "shipped",
        "Objeto recebido na unidade de distribuição",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDI", 45): (
        "shipped",
        "Objeto recebido na unidade de distribuição",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDR", 45): (
        "shipped",
        "Objeto recebido na unidade de distribuição",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDE", 46): (
        "delivery_unsuccessful",
        "Tentativa de entrega não efetuada",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDI", 46): (
        "delivery_unsuccessful",
        "Tentativa de entrega não efetuada",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDR", 46): (
        "delivery_unsuccessful",
        "Tentativa de entrega não efetuada",
        "Entrega prevista para o próximo dia útil",
        "Acompanhar",
    ),
    ("BDE", 47): (
        "delivery_unsuccessful",
        "Saída para entrega cancelada",
        "Será efetuado novo lançamento para entrega",
        "Acompanhar",
    ),
    ("BDI", 47): (
        "delivery_unsuccessful",
        "Saída para entrega cancelada",
        "Será efetuado novo lançamento para entrega",
        "Acompanhar",
    ),
    ("BDR", 47): (
        "delivery_unsuccessful",
        "Saída para entrega cancelada",
        "Será efetuado novo lançamento para entrega",
        "Acompanhar",
    ),
    ("BDE", 48): (
        "delivery_rejected",
        "Retirada em Unidade dos Correios não autorizada pelo remetente",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 48): (
        "delivery_rejected",
        "Retirada em Unidade dos Correios não autorizada pelo remetente",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 48): (
        "delivery_rejected",
        "Retirada em Unidade dos Correios não autorizada pelo remetente",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 49): (
        "delivery_unsuccessful",
        "As dimensões do objeto impossibilitam o tratamento e a entrega",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 49): (
        "delivery_unsuccessful",
        "As dimensões do objeto impossibilitam o tratamento e a entrega",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 49): (
        "delivery_unsuccessful",
        "As dimensões do objeto impossibilitam o tratamento e a entrega",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 50): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 50): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 50): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 51): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 51): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 51): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 52): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDI", 52): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 52): (
        "lost",
        "Objeto roubado",
        "Favor entrar em contato com os Correios.",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 53): (
        "shipped",
        "Objeto reimpresso e reenviado",
        "",
        "Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.",
    ),
    ("BDI", 53): (
        "shipped",
        "Objeto reimpresso e reenviado",
        "",
        "Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.",
    ),
    ("BDR", 53): (
        "shipped",
        "Objeto reimpresso e reenviado",
        "",
        "Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.",
    ),
    ("BDE", 54): (
        "customs_control",
        "Para recebimento do objeto, é necessário o pagamento do ICMS Importação",
        "",
        "Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.",
    ),
    ("BDI", 54): (
        "customs_control",
        "Para recebimento do objeto, é necessário o pagamento do ICMS Importação",
        "",
        "Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.",
    ),
    ("BDR", 54): (
        "customs_control",
        "Para recebimento do objeto, é necessário o pagamento do ICMS Importação",
        "",
        "Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.",
    ),
    ("BDE", 55): ("customs_control", "Solicitada revisão do tributo estabelecido", "", "Acompanhar"),
    ("BDI", 55): ("customs_control", "Solicitada revisão do tributo estabelecido", "", "Acompanhar"),
    ("BDR", 55): ("customs_control", "Solicitada revisão do tributo estabelecido", "", "Acompanhar"),
    ("BDE", 56): (
        "customs_control",
        "Declaração aduaneira ausente ou incorreta",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDI", 56): (
        "customs_control",
        "Declaração aduaneira ausente ou incorreta",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDR", 56): (
        "customs_control",
        "Declaração aduaneira ausente ou incorreta",
        "Objeto será devolvido ao remetente",
        "Acompanhar o retorno do objeto ao remetente.",
    ),
    ("BDE", 57): ("customs_control", "Revisão de tributo concluída - Objeto liberado", "", "Acompanhar"),
    ("BDI", 57): ("customs_control", "Revisão de tributo concluída - Objeto liberado", "", "Acompanhar"),
    ("BDR", 57): ("customs_control", "Revisão de tributo concluída - Objeto liberado", "", "Acompanhar"),
    ("BDE", 58): (
        "customs_control",
        "Revisão de tributo concluída - Tributo alterado",
        "O valor do tributo pode ter aumentado ou diminuído",
        "Acompanhar",
    ),
    ("BDI", 58): (
        "customs_control",
        "Revisão de tributo concluída - Tributo alterado",
        "O valor do tributo pode ter aumentado ou diminuído",
        "Acompanhar",
    ),
    ("BDR", 58): (
        "customs_control",
        "Revisão de tributo concluída - Tributo alterado",
        "O valor do tributo pode ter aumentado ou diminuído",
        "Acompanhar",
    ),
    ("BDE", 59): (
        "customs_control",
        "Revisão de tributo concluída - Tributo mantido",
        "Poderá haver incidência de juros e multa.",
        "Acompanhar",
    ),
    ("BDI", 59): (
        "customs_control",
        "Revisão de tributo concluída - Tributo mantido",
        "Poderá haver incidência de juros e multa.",
        "Acompanhar",
    ),
    ("BDR", 59): (
        "customs_control",
        "Revisão de tributo concluída - Tributo mantido",
        "Poderá haver incidência de juros e multa.",
        "Acompanhar",
    ),
    ("BDR", 60): ("shipped", "O objeto encontra-se aguardando prazo para refugo", "", "Acompanhar"),
    ("BDI", 60): ("shipped", "O objeto encontra-se aguardando prazo para refugo", "", "Acompanhar"),
    ("BDE", 66): (
        "shipped_delayed",
        "Área com distribuição sujeita a prazo diferenciado",
        "Restrição de entrega domiciliar temporária",
        "Acompanhar",
    ),
    ("BDI", 66): (
        "shipped_delayed",
        "Área com distribuição sujeita a prazo diferenciado",
        "Restrição de entrega domiciliar temporária",
        "Acompanhar",
    ),
    ("BDR", 66): (
        "shipped_delayed",
        "Área com distribuição sujeita a prazo diferenciado",
        "Restrição de entrega domiciliar temporária",
        "Acompanhar",
    ),
    ("BDE", 68): (
        "waiting_retrieval",
        "Objeto entregue na Caixa de Correios Inteligente",
        "Estará disponível por até 3 dias, a partir de hoje",
        "Acompanhar",
    ),
    ("BDI", 68): (
        "waiting_retrieval",
        "Objeto entregue na Caixa de Correios Inteligente",
        "Estará disponível por até 3 dias, a partir de hoje",
        "Acompanhar",
    ),
    ("BDR", 68): (
        "waiting_retrieval",
        "Objeto entregue na Caixa de Correios Inteligente",
        "Estará disponível por até 3 dias, a partir de hoje",
        "Acompanhar",
    ),
    ("BDE", 69): ("shipped_delayed", "Objeto com atraso na entrega", "", "Acompanhar"),
    ("BDI", 69): ("shipped_delayed", "Objeto com atraso na entrega", "", "Acompanhar"),
    ("BDR", 69): ("shipped_delayed", "Objeto com atraso na entrega", "", "Acompanhar"),
    ("BDE", 70): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDR", 70): (
        "delivered",
        "Objeto entregue ao destinatário",
        "Recebido por:",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDE", 71): (
        "customs_control",
        "Objeto apreendido por: ANVISA/VIGILANCIA SANITARIA",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 71): (
        "customs_control",
        "Objeto apreendido por: ANVISA/VIGILANCIA SANITARIA",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 72): (
        "customs_control",
        "Objeto apreendido por: BOMBEIROS",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 72): (
        "customs_control",
        "Objeto apreendido por: BOMBEIROS",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 73): (
        "customs_control",
        "Objeto apreendido por: EXERCITO",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 73): (
        "customs_control",
        "Objeto apreendido por: EXERCITO",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 74): (
        "customs_control",
        "Objeto apreendido por: IBAMA",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 74): (
        "customs_control",
        "Objeto apreendido por: IBAMA",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 75): (
        "customs_control",
        "Objeto apreendido por: POLICIA FEDERAL",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 75): (
        "customs_control",
        "Objeto apreendido por: POLICIA FEDERAL",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 76): (
        "customs_control",
        "Objeto apreendido por: POLICIA CIVIL/MILITAR",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDR", 76): (
        "customs_control",
        "Objeto apreendido por: POLICIA CIVIL/MILITAR",
        "O objeto está em poder da autoridade competente",
        "Acionar atendimento dos Correios.",
    ),
    ("BDE", 79): (
        "delivered",
        "Objeto entregue ao contratante",
        "Contingência COVID-19",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDI", 79): (
        "delivered",
        "Objeto entregue ao contratante",
        "Contingência COVID-19",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDR", 79): (
        "delivered",
        "Objeto entregue ao contratante",
        "Contingência COVID-19",
        "Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.",
    ),
    ("BDE", 80): ("lost", "Objeto extraviado", "", "Acionar a CAC dos Correios"),
    ("BDI", 80): ("lost", "Objeto extraviado", "", "Acionar a CAC dos Correios"),
    ("BDR", 80): ("lost", "Objeto extraviado", "", "Acionar a CAC dos Correios"),
    ("BDE", 89): ("lost", "Localidade desconhecida no Brasil", "Objeto será devolvido ao remetente", "Acompanhar"),
    ("BDI", 89): ("lost", "Localidade desconhecida no Brasil", "Objeto será devolvido ao remetente", "Acompanhar"),
    ("BDR", 89): ("lost", "Localidade desconhecida no Brasil", "Objeto será devolvido ao remetente", "Acompanhar"),
    ("CMR", 1): ("shipped", "Conferido", "", "Acompanhar"),
}

ZIP_CODE_MAP = {
    "SP": (RangeSet((1000, 20000)), RangeSet((1000, 6000), (8000, 8500))),  # all  # capital
    "RJ": (RangeSet((20000, 29000)), RangeSet((20000, 23800))),
    "ES": (RangeSet((29000, 30000)), RangeSet((29000, 29100))),
    "MG": (RangeSet((30000, 40000)), RangeSet((30000, 32000))),
    "BA": (RangeSet((40000, 49000)), RangeSet((40000, 42600))),
    "SE": (RangeSet((49000, 50000)), RangeSet((49000, 49100))),
    "PE": (RangeSet((50000, 57000)), RangeSet((50000, 53000))),
    "AL": (RangeSet((57000, 58000)), RangeSet((57000, 57100))),
    "PB": (RangeSet((58000, 59000)), RangeSet((58000, 58100))),
    "RN": (RangeSet((59000, 60000)), RangeSet((59000, 59140))),
    "CE": (RangeSet((60000, 64000)), RangeSet((60000, 61600))),
    "PI": (RangeSet((64000, 65000)), RangeSet((64000, 64100))),
    "MA": (RangeSet((65000, 66000)), RangeSet((65000, 65110))),
    "PA": (RangeSet((66000, 68900)), RangeSet((66000, 67000))),
    "AP": (RangeSet((68900, 69000)), RangeSet((68900, 68915))),
    "AM": (RangeSet((69000, 69300), (69400, 69900)), RangeSet((69000, 69100), (69000, 69100))),
    "RR": (RangeSet((69300, 69400)), RangeSet((69300, 69340))),
    "AC": (RangeSet((69900, 70000)), RangeSet((69900, 69925))),
    "DF": (RangeSet((70000, 72800), (73000, 73700)), RangeSet((70000, 72800), (73000, 73700))),
    "GO": (RangeSet((72800, 73000), (73700, 76800)), RangeSet((74000, 74900))),
    "TO": (RangeSet((77000, 78000)), RangeSet((77000, 77300))),
    "MT": (RangeSet((78000, 78900)), RangeSet((78000, 78110))),
    "RO": (RangeSet((76800, 77000), (78900, 79000)), RangeSet((76800, 76850), (78900, 78925))),
    "MS": (RangeSet((79000, 80000)), RangeSet((79000, 79130))),
    "PR": (RangeSet((80000, 88000)), RangeSet((80000, 83000))),
    "SC": (RangeSet((88000, 90000)), RangeSet((88000, 88100))),
    "RS": (RangeSet((90000, 100000)), RangeSet((90000, 92000))),
}  # type: Dict[str, Tuple[RangeSet, RangeSet]]

ZIP_CODES = RangeSet(*(p[0] for p in ZIP_CODE_MAP.values()))

FREIGHT_ERROR_RESPONSES = {
    0: "Processamento com sucesso",
    -1: "Código de serviço inválido",
    -2: "CEP de origem inválido",
    -3: "CEP de destino inválido",
    -4: "Peso excedido",
    -5: "O Valor Declarado não deve exceder R$ 10.000,00",
    -6: "Serviço indisponível para o trecho informado",
    -7: "O Valor Declarado é obrigatório para este serviço",
    -8: "Este serviço não aceita Mão Própria",
    -9: "Este serviço não aceita Aviso de Recebimento",
    -10: "Precificação indisponível para o trecho informado",
    -11: "Para definição do preço deverão ser informados, também, o comprimento, "
    "a largura e altura do objeto em centímetros (cm).",
    -12: "Comprimento inválido.",
    -13: "Largura inválida.",
    -14: "Altura inválida.",
    -15: "O comprimento não pode ser maior que 100 cm.",
    -16: "A largura não pode ser maior que 100 cm.",
    -17: "A altura não pode ser maior que 100 cm.",
    -18: "A altura não pode ser inferior a 1 cm.",
    -20: "A largura não pode ser inferior a 10 cm.",
    -22: "O comprimento não pode ser inferior a 15 cm.",
    -23: "A soma resultante do comprimento + largura + altura não deve superar a 200 cm.",
    -24: "Comprimento inválido.",
    -25: "Diâmetro inválido",
    -26: "Informe o comprimento.",
    -27: "Informe o diâmetro.",
    -28: "O comprimento não pode ser maior que 100 cm.",
    -29: "O diâmetro não pode ser maior que 91 cm.",
    -30: "O comprimento não pode ser inferior a 18 cm.",
    -31: "O diâmetro não pode ser inferior a 5 cm.",
    -32: "A soma resultante do comprimento + o dobro do diâmetro não deve superar a 200 cm.",
    -33: "Sistema temporariamente fora do ar. Favor tentar mais tarde.",
    -34: "Código Administrativo ou Senha inválidos.",
    -35: "Senha incorreta.",
    -36: "Cliente não possui contrato vigente com os Correios.",
    -37: "Cliente não possui serviço ativo em seu contrato.",
    -38: "Serviço indisponível para este código administrativo.",
    -39: "Peso excedido para o formato envelope",
    -40: (
        "Para definicao do preco deverao ser informados, tambem, o comprimento e a "
        "largura e altura do objeto em centimetros (cm)."
    ),
    -41: "O comprimento nao pode ser maior que 60 cm.",
    -42: "O comprimento nao pode ser inferior a 15 cm.",
    -43: "A soma resultante do comprimento + largura nao deve superar a 120 cm.",
    -44: "A largura nao pode ser inferior a 10 cm.",
    -45: "A largura nao pode ser maior que 60 cm.",
    -888: "Erro ao calcular a tarifa",
    6: "Localidade de origem não abrange o serviço informado",
    # 7: 'Localidade de destino não abrange o serviço informado',
    7: "Serviço indisponível, tente mais tarde",  #
    8: "Serviço indisponível para o trecho informado",
    9: "CEP inicial pertencente a Área de Risco.",
    10: "Área com entrega temporariamente sujeita a prazo diferenciado.",
    11: "CEP inicial e final pertencentes a Área de Risco",
    99: "Outros erros diversos do .Net",
}

FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED = 9
FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED = 10
FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED = 11



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from unittest import mock

import pytest

from correios.utils import RangeSet, capitalize_phrase, get_resource_path, rreplace, to_decimal, to_integer

phrase = "FOo bAr BAZ qux"


@pytest.fixture
def rangeset():
    """ returns the sequence 1, 2, 4, 5, 7, 8 """
    return RangeSet((1, 3), (4, 6), (7, 9))


def test_rangeset():
    assert list(RangeSet(range(3))) == [0, 1, 2]
    assert list(RangeSet((0, 3))) == [0, 1, 2]
    assert list(RangeSet((1, 3), (4, 6), (7, 9))) == [1, 2, 4, 5, 7, 8]
    assert list(RangeSet((1, 3), RangeSet((4, 6)), range(7, 9))) == [1, 2, 4, 5, 7, 8]

    with pytest.raises(ValueError):
        RangeSet(1)


@pytest.mark.parametrize("ranges_len, elements_len, ranges", ((1, 3, (range(3),)), (3, 6, ((1, 3), (4, 6), (7, 9)))))
def test_rangeset_len(ranges_len, elements_len, ranges):
    rangeset = RangeSet(*ranges)
    assert len(rangeset) == elements_len
    assert len(rangeset.ranges) == ranges_len


@pytest.mark.parametrize("element", (1, 2, 4, 5, 7, 8))
def test_rangeset_contain(element, rangeset):
    assert element in rangeset


@pytest.mark.parametrize("element", (-1, 3, 6, 9, 10000, 12))
def test_rangeset_does_not_contain(element, rangeset):
    assert element not in rangeset


def test_rangeset_iter(rangeset):
    assert list(rangeset) == [1, 2, 4, 5, 7, 8]


@pytest.mark.parametrize("phrase", (phrase, phrase.upper(), phrase.lower()))
def test_capitalize_phrase(phrase):
    assert capitalize_phrase(phrase) == "Foo Bar Baz Qux"


def test_rreplace():
    phrase = "foo bar baz qux"
    assert rreplace(phrase, " ", "-", 1) == "foo bar baz-qux"
    assert rreplace(phrase, " ", "-", 2) == "foo bar-baz-qux"
    assert rreplace(phrase, " ", "-", 3) == "foo-bar-baz-qux"
    assert rreplace(phrase, " ", "-") == "foo-bar-baz-qux"


@pytest.mark.parametrize(
    "s, d",
    (
        ("", Decimal("0.00")),
        ("3", Decimal("3.00")),
        ("3.57", Decimal("3.57")),
        ("3.468", Decimal("3.47")),
        ("3.4", Decimal("3.40")),
        ("3,57", Decimal("3.57")),
        ("3,468", Decimal("3.47")),
        ("3,4", Decimal("3.40")),
        ("1,357.93", Decimal("1357.93")),
        ("1.357,93", Decimal("1357.93")),
        ("1_357.93", Decimal("1357.93")),
        ("1_357,93", Decimal("1357.93")),
    ),
)
def test_to_decimal(s, d):
    assert to_decimal(s) == d


@pytest.mark.parametrize(
    "v, p, r",
    (
        ("3.4", 1, Decimal("3.4")),
        ("3.4", 4, Decimal("3.4000")),
        ("3.4", 0, Decimal("3")),
        ("3.6", 0, Decimal("4")),
        ("3.46876", 2, Decimal("3.47")),
    ),
)
def test_to_decimal_precision(v, p, r):
    assert to_decimal(v, p) == r


@pytest.mark.parametrize("v, r", ((3, 3), ("3", 3), ("  \t3  \n", 3)))
def test_to_integer(v, r):
    assert to_integer(v) == r


def test_get_wsdl_file_path():
    path = get_resource_path("fake")
    assert str(path).endswith("correios/data/fake")


@mock.patch("pkg_resources.resource_filename", return_value="/")
def test_should_use_pkg_resources_to_get_wsdl_files(mock_resource):
    path = get_resource_path("fake")

    mock_resource.assert_called_with("correios", "data/fake")
    assert str(path) == "/"



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from typing import Dict, Optional

from ..utils import to_decimal, to_integer
from .address import ReceiverAddress, SenderAddress, ZipAddress
from .posting import (
    EventStatus,
    FreightResponse,
    NotFoundTrackingEvent,
    Package,
    PostalUnit,
    PostInfo,
    PostingList,
    Receipt,
    ShippingLabel,
    TrackingCode,
    TrackingEvent,
)
from .user import Contract, ExtraService, FederalTaxNumber, PostingCard, Service, StateTaxNumber, User


class ModelBuilder:
    def build_service(self, service_data):
        service = Service(
            code=service_data.codigo,
            id=service_data.id,
            description=service_data.descricao,
            category=service_data.servicoSigep.categoriaServico,
        )
        return service

    def build_posting_card(self, contract: Contract, posting_card_data):
        posting_card = PostingCard(
            contract=contract,
            number=posting_card_data.numero,
            administrative_code=posting_card_data.codigoAdministrativo,
        )

        posting_card.start_date = posting_card_data.dataVigenciaInicio
        posting_card.end_date = posting_card_data.dataVigenciaFim
        posting_card.status = posting_card_data.statusCartaoPostagem
        posting_card.status_code = posting_card_data.statusCodigo
        posting_card.unit = posting_card_data.unidadeGenerica.strip() or 0

        for service_data in posting_card_data.servicos:
            service = self.build_service(service_data)
            posting_card.add_service(service)

        return posting_card

    def build_contract(self, user: User, contract_data):
        contract = Contract(
            user=user, number=contract_data.contratoPK.numero, regional_direction=contract_data.codigoDiretoria
        )

        contract.customer_code = contract_data.codigoCliente
        contract.status_code = contract_data.statusCodigo
        contract.start_date = contract_data.dataVigenciaInicio
        contract.end_date = contract_data.dataVigenciaFim

        for posting_card_data in contract_data.cartoesPostagem:
            self.build_posting_card(contract, posting_card_data)

        return contract

    def build_user(self, user_data):
        user = User(
            name=user_data.nome,
            federal_tax_number=FederalTaxNumber(user_data.cnpj),
            state_tax_number=StateTaxNumber(user_data.inscricaoEstadual),
            status_number=user_data.statusCodigo,
        )

        for contract_data in user_data.contratos:
            self.build_contract(user, contract_data)

        return user

    def build_zip_address(self, zip_address_data):
        zip_address = ZipAddress(
            id=0,
            zip_code=zip_address_data.cep,
            state=zip_address_data.uf,
            city=zip_address_data.cidade,
            district=zip_address_data.bairro,
            address=zip_address_data.end,
            complements=[zip_address_data.complemento2],
        )
        return zip_address

    def build_post_info(self, data, user: User) -> PostInfo:
        post_info = PostInfo(
            postal_unit=self.build_postal_unit(data.plp),
            posting_list=self._load_posting_list(data=data, user=user),
            value=data.plp.valor_global,
        )
        return post_info

    def build_receipt(self, data) -> Optional[Receipt]:
        if data.status_processamento != Receipt.STATUS_PROCESSED:
            return None

        receipt = Receipt(
            number=data.numero_comprovante_postagem,
            post_date=data.data_postagem_sara.text,
            value=data.valor_cobrado.text,
        )
        return receipt

    def build_postal_unit(self, data) -> PostalUnit:
        postal_unit = PostalUnit(code=data.mcu_unidade_postagem, description=data.nome_unidade_postagem)
        return postal_unit

    def _load_posting_list(self, data, user: User) -> PostingList:
        contract_number = to_integer(data.remetente.numero_contrato)

        contract = next(c for c in user.contracts if c.number == contract_number)

        posting_card_number = str(data.plp.cartao_postagem.text)

        posting_card = next(p for p in contract.posting_cards if p.number == posting_card_number)

        posting_list = PostingList(custom_id=0)

        for postal_object in data.objeto_postal:

            posting_list.add_shipping_label(
                self._load_shipping_label(
                    data=postal_object,
                    posting_card=posting_card,
                    sender_address=self._load_sender_address(data.remetente),
                )
            )

        posting_list.close_with_id(data.plp.id_plp)

        return posting_list

    def _load_shipping_label(self, data, posting_card: PostingCard, sender_address: SenderAddress) -> ShippingLabel:

        declared_value = getattr(data.servico_adicional, "valor_declarado", None)
        extra_services_codes = list(data.servico_adicional.codigo_servico_adicional)
        extra_services = [ExtraService.get(code) for code in extra_services_codes if code]

        invoice_value = getattr(data.nacional, "valor_nota_fiscal", None)

        billing = getattr(data.nacional, "valor_a_cobrar", None) or "0.00"

        shipping_label = ShippingLabel(
            billing=to_decimal(billing),
            invoice_number=data.nacional.numero_nota_fiscal,
            invoice_series=data.nacional.serie_nota_fiscal,
            value=to_decimal(declared_value or invoice_value or "0.00"),
            text=data.nacional.descricao_objeto,
            posting_card=posting_card,
            sender=sender_address,
            receiver=self._load_receiver_address(data),
            package=self._load_package(data),
            service=Service.get(data.codigo_servico_postagem.text),
            tracking_code=data.numero_etiqueta.text,
            receipt=self.build_receipt(data),
        )

        shipping_label.add_extra_services(
            [extra_service for extra_service in extra_services if extra_service not in shipping_label.extra_services]
        )

        return shipping_label

    def _load_sender_address(self, data) -> SenderAddress:
        sender_address = SenderAddress(
            email=data.email_remetente.text,
            name=data.nome_remetente.text,
            street=data.logradouro_remetente.text,
            number=data.numero_remetente.text,
            complement=data.complemento_remetente.text,
            neighborhood=data.bairro_remetente.text,
            zip_code=data.cep_remetente.text,
            city=data.cidade_remetente.text,
            state=data.uf_remetente.text,
            phone=data.telefone_remetente.text or "",
        )

        return sender_address

    def _load_receiver_address(self, data) -> ReceiverAddress:
        receiver_data = data.destinatario

        extra_data = data.nacional

        receiver_address = ReceiverAddress(
            email=getattr(receiver_data, "email_remetente", ""),
            name=receiver_data.nome_destinatario.text or "",
            street=receiver_data.logradouro_destinatario.text or "",
            number=receiver_data.numero_end_destinatario.text or "",
            complement=receiver_data.complemento_destinatario.text or "",
            neighborhood=extra_data.bairro_destinatario.text or "",
            zip_code=extra_data.cep_destinatario.text or "",
            city=extra_data.cidade_destinatario.text or "",
            state=extra_data.uf_destinatario.text or "",
            phone=receiver_data.celular_destinatario.text or "",
        )

        return receiver_address

    def _build_package_data(self, data):
        dimensions = data.dimensao_objeto

        if dimensions.tipo_objeto == Package.TYPE_BOX:
            package_data = {
                "height": dimensions.dimensao_altura.text,
                "length": dimensions.dimensao_comprimento.text,
                "weight": data.peso.text,
                "width": dimensions.dimensao_largura.text,
            }
        elif dimensions.tipo_objeto == Package.TYPE_CYLINDER:
            package_data = {
                "diameter": dimensions.dimensao_diametro.text,
                "length": dimensions.dimensao_comprimento.text,
                "weight": data.peso.text,
            }
        else:
            package_data = {"weight": data.peso.text}

        package_data = {k: float(v.replace(",", ".")) for (k, v) in package_data.items()}
        package_data["package_type"] = dimensions.tipo_objeto
        return package_data

    def _load_package(self, data) -> Package:
        package_data = self._build_package_data(data)
        package = Package(service=data.codigo_servico_postagem.text, **package_data)
        return package

    def build_posting_card_status(self, response):
        if response.lower() != "normal":
            return PostingCard.CANCELLED
        return PostingCard.ACTIVE

    def build_tracking_codes_list(self, response):
        codes = response.split(",")
        return TrackingCode.create_range(codes[0], codes[1])

    def _load_invalid_event(self, tracking_code: TrackingCode, tracked_object):
        event = NotFoundTrackingEvent(timestamp=datetime.utcnow(), comment=tracked_object.erro)
        tracking_code.add_event(event)

    def _load_events(self, tracking_code: TrackingCode, events):
        for event in events:
            timestamp = datetime.strptime("{} {}".format(event.data, event.hora), TrackingEvent.timestamp_format)
            event = TrackingEvent(
                timestamp=timestamp,
                status=EventStatus(event.tipo, event.status),
                location_zip_code=getattr(event, "codigo", ""),
                location=getattr(event, "local", ""),
                city=getattr(event, "cidade", ""),
                state=getattr(event, "uf", ""),
                receiver=getattr(event, "recebedor", ""),
                document=getattr(event, "documento", ""),
                comment=getattr(event, "comentario", ""),
                description=getattr(event, "descricao", ""),
                details=getattr(event, "detalhes", ""),
            )

            tracking_code.add_event(event)

    def load_tracking_events(self, tracking_codes: Dict[str, TrackingCode], response):
        result = []
        for tracked_object in response.objeto:
            tracking_code = tracking_codes[tracked_object.numero]

            if "erro" in tracked_object and tracked_object.erro:
                self._load_invalid_event(tracking_code, tracked_object)
            else:
                tracking_code.name = tracked_object.nome
                tracking_code.initials = tracked_object.sigla
                tracking_code.category = tracked_object.categoria
                self._load_events(tracking_code, tracked_object.evento)

            result.append(tracking_code)

        return result

    def build_freights_list(self, response):
        result = []
        for service_data in response.cServico:
            freight = self.build_freight(service_data=service_data)
            result.append(freight)
        return result

    def build_freight(self, service_data) -> FreightResponse:
        return FreightResponse(
            service=Service.get(service_data.Codigo),
            delivery_time=int(service_data.PrazoEntrega),
            value=to_decimal(service_data.ValorSemAdicionais),
            declared_value=to_decimal(service_data.ValorValorDeclarado),
            mp_value=to_decimal(service_data.ValorMaoPropria),
            ar_value=to_decimal(service_data.ValorAvisoRecebimento),
            saturday=service_data.EntregaSabado and service_data.EntregaSabado.lower() == "s" or False,
            home=service_data.EntregaDomiciliar and service_data.EntregaDomiciliar.lower() == "s" or False,
            error_code=to_integer(service_data.Erro) or 0,
            error_message=service_data.MsgErro or "",
        )



from unittest import mock

import pytest

from .vcr import replace_auth, vcr


@pytest.fixture
def username():
    return "cuca"


@pytest.fixture
def password():
    return "beludo"


@pytest.fixture
def request_body(username, password):
    body = (
        "<?xml version=''1.0'' encoding=''utf-8''?>"
        '<soap-env:Envelope xmlns:soap-env="http://schemas.xmlsoap.org/soap/envelope/">'
        '<soap-env:Body><ns0:buscaEventosLista xmlns:ns0="http://resource.webservice.correios.com.br/">'
        "<usuario>{}</usuario><senha>{}senha><tipo>L</tipo><resultado>T</resultado><lingua>101</lingua>"
        "<objetos>JB683971943BR</objetos><objetos>JT365572014BR</objetos>"
        "</ns0:buscaEventosLista></soap-env:Body></soap-env:Envelope>"
        ""
    )
    body = body.format(username, password)
    return body.encode()


def test_replace_auth(request_body, username, password):
    request = mock.Mock(body=request_body)
    body = str(replace_auth(request))

    assert username not in body
    assert password not in body


def test_vcr_uses_replace_auth():
    assert vcr.before_record_request == replace_auth



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from unittest import mock

import pytest
from requests.exceptions import ConnectTimeout
from zeep.exceptions import Fault

from correios.exceptions import (
    AuthenticationError,
    CanceledPostingCardError,
    ClientError,
    ClosePostingListError,
    ConnectTimeoutError,
    NonexistentPostingCardError,
    PostingListSerializerError,
    TrackingCodesLimitExceededError,
)
from correios.models.address import ZipCode
from correios.models.builders import ModelBuilder
from correios.models.data import (
    EXTRA_SERVICE_AR,
    EXTRA_SERVICE_MP,
    EXTRA_SERVICE_VD_PAC,
    EXTRA_SERVICE_VD_PAC_MINI,
    EXTRA_SERVICE_VD_SEDEX,
    FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED,
    FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED,
    SERVICE_PAC,
    SERVICE_SEDEX,
    SERVICE_SEDEX10,
)
from correios.models.posting import (
    FreightResponse,
    NotFoundTrackingEvent,
    PostalUnit,
    PostInfo,
    PostingList,
    TrackingCode,
)
from correios.models.user import ExtraService, PostingCard, Service
from correios.serializers import PostingListSerializer
from correios.utils import get_resource_path, to_decimal
from correios.xml_utils import fromstring

from .vcr import vcr

try:
    from correios import client as correios
except ImportError:
    correios = None


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_basic_client():
    client = correios.Correios(username="sigep", password="XXXXXX", environment=correios.Correios.TEST)
    assert client.sigep_url == str(get_resource_path("wsdls/AtendeCliente-test.wsdl"))
    assert not client.sigep_verify
    assert client.username == "sigep"
    assert client.password == "XXXXXX"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@mock.patch("zeep.proxy.OperationProxy.__call__")
def test_client_timeout_error(mock_soap_client, client):
    mock_soap_client.side_effect = ConnectTimeout()
    with pytest.raises(ConnectTimeoutError):
        client.find_zipcode(ZipCode("70002-900"))


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_authentication_error(client):
    with pytest.raises(AuthenticationError):
        client.get_user(contract_number="9911222777", posting_card_number="0056789123")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_canceled_posting_card_error(client):
    with pytest.raises(CanceledPostingCardError):
        client.get_user(contract_number="9911222777", posting_card_number="0057018901")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_client_nonexistent_posting_card_error(client):
    with pytest.raises(NonexistentPostingCardError):
        client.get_user(contract_number="9911222777", posting_card_number="4444444444")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_user(client):
    user = client.get_user(contract_number="9911222777", posting_card_number="0056789123")

    assert user.name == "ECT"
    assert user.federal_tax_number == "34028316000103"
    assert user.state_tax_number == "0733382100116"
    assert user.status_number == 1
    assert len(user.contracts) == 1

    contract = user.contracts[0]
    assert len(contract.posting_cards) == 1


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_find_zip_code(client):
    zip_address = client.find_zipcode(ZipCode("70002-900"))

    assert zip_address.id == 0
    assert zip_address.zip_code == "70002900"
    assert zip_address.state == "DF"
    assert zip_address.city == "Brasília"
    assert zip_address.district == "Asa Norte"
    assert zip_address.address == "SBN Quadra 1 Bloco A"
    assert zip_address.complements == []


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_verify_service_availability(client, posting_card):
    status = client.verify_service_availability(posting_card, SERVICE_SEDEX10, "82940150", "01310000")
    assert status is True


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_posting_card_status(client, posting_card):
    status = client.get_posting_card_status(posting_card)
    assert status == PostingCard.ACTIVE


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_request_tracking_codes(client, user):
    result = client.request_tracking_codes(user, Service.get(SERVICE_SEDEX), quantity=10)
    assert len(result) == 10
    assert len(result[0].code) == 13


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_generate_verification_digit(client):
    result = client.generate_verification_digit(["DL74668653 BR"])
    assert result[0] == 6


@vcr.use_cassette
def test_get_post_info(client):
    result = client._auth_call("solicitaXmlPlp", 875057)

    data = fromstring(result.encode("iso-8859-1"))

    user = client.get_user(
        contract_number=data.remetente.numero_contrato.text, posting_card_number=data.plp.cartao_postagem.text
    )

    post_info = client.get_post_info(number=875057)
    assert isinstance(post_info, PostInfo)
    assert str(post_info.value) == data.plp.valor_global.text

    postal_unit = post_info.postal_unit
    assert isinstance(postal_unit, PostalUnit)
    assert postal_unit.code == data.plp.mcu_unidade_postagem
    assert postal_unit.description == data.plp.nome_unidade_postagem

    posting_list = post_info.posting_list
    assert isinstance(posting_list, PostingList)
    assert posting_list.number == data.plp.id_plp

    shipping_labels = posting_list.shipping_labels
    assert len(shipping_labels) == len(data.objeto_postal)

    for obj in data.objeto_postal:
        tracking_code = TrackingCode.create(obj.numero_etiqueta.text)
        assert tracking_code.short in shipping_labels
        label = shipping_labels[tracking_code.short]

        extra_info = obj.nacional
        billing = getattr(extra_info, "valor_a_cobrar", None) or "0.00"
        assert label.billing == to_decimal(billing)
        assert label.invoice_number == extra_info.numero_nota_fiscal
        assert label.invoice_series == extra_info.serie_nota_fiscal

        extra_services = obj.servico_adicional

        declared_value = getattr(extra_services, "valor_declarado", None)

        invoice_value = getattr(extra_info, "valor_nota_fiscal", None)

        assert label.real_value == to_decimal(declared_value or invoice_value or "0.00")

        assert label.text == extra_info.descricao_objeto
        posting_card_number = user.contracts[0].posting_cards[0].number
        assert label.posting_card.number == posting_card_number

        sender = data.remetente

        assert label.sender.email == sender.email_remetente.text
        assert label.sender.name == sender.nome_remetente.text
        assert label.sender.street == sender.logradouro_remetente.text
        assert label.sender.number == sender.numero_remetente.text
        assert label.sender.complement == sender.complemento_remetente.text
        assert label.sender.neighborhood == sender.bairro_remetente.text
        assert label.sender.zip_code == sender.cep_remetente.text
        assert label.sender.city == sender.cidade_remetente.text
        assert label.sender.state == sender.uf_remetente.text
        assert label.sender.phone.number == (sender.telefone_remetente.text or "")

        receiver = obj.destinatario

        assert label.receiver.email == (receiver.email_destinatario.text or "")

        assert label.receiver.name == receiver.nome_destinatario.text
        assert label.receiver.street == receiver.logradouro_destinatario.text
        assert label.receiver.number == receiver.numero_end_destinatario.text
        assert label.receiver.complement == (receiver.complemento_destinatario.text or "")
        assert label.receiver.neighborhood == extra_info.bairro_destinatario.text
        assert label.receiver.zip_code == extra_info.cep_destinatario.text
        assert label.receiver.city == extra_info.cidade_destinatario.text
        assert label.receiver.state == extra_info.uf_destinatario.text
        assert label.receiver.phone.number == (receiver.celular_destinatario.text or "")

        assert len(label.extra_services) == len(extra_services)

        for service in extra_services.codigo_servico_adicional:
            assert service in label.extra_services

        package = label.package

        assert package.service == obj.codigo_servico_postagem

        dimensions = obj.dimensao_objeto

        assert package.package_type == dimensions.tipo_objeto

        assert package.real_diameter == float(dimensions.dimensao_diametro.text.replace(",", "."))

        assert package.real_height == float(dimensions.dimensao_altura.text.replace(",", "."))

        assert package.real_length == float(dimensions.dimensao_comprimento.text.replace(",", "."))

        assert package.real_weight == float(obj.peso.text.replace(",", "."))
        assert package.real_width == float(dimensions.dimensao_largura.text.replace(",", "."))

        receipt = label.receipt

        assert receipt.number == obj.numero_comprovante_postagem
        assert receipt.real_post_date == obj.data_postagem_sara.text
        assert receipt.real_value == obj.valor_cobrado.text


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_close_posting_list(client, posting_card, posting_list, shipping_label):
    shipping_label.posting_card = posting_card
    posting_list.add_shipping_label(shipping_label)
    posting_list = client.close_posting_list(posting_list, posting_card)
    assert posting_list.number is not None
    assert posting_list.closed


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@mock.patch("zeep.proxy.OperationProxy.__call__")
def test_client_close_posting_list_error(mock_soap_client, client, posting_card, posting_list, shipping_label):
    original_exception = Fault(
        "A PLP não será fechada , o(s) objeto(s) [PP40233163BR] já estão vinculados em outra PLP!"
    )
    mock_soap_client.side_effect = original_exception
    shipping_label.posting_card = posting_card
    posting_list.add_shipping_label(shipping_label)

    with pytest.raises(ClosePostingListError) as exc:
        client.close_posting_list(posting_list, posting_card)

    assert isinstance(exc.value.__cause__, ClientError)
    assert str(exc.value.__cause__) == original_exception.message


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_codes_events(client):
    codes = ["BE058714266BR", "JT365572014BR"]
    result = client.get_tracking_code_events(codes)

    assert len(result) == 2
    assert result[0] != result[1]

    assert isinstance(result[0], TrackingCode)
    assert result[0].code in codes

    assert isinstance(result[1], TrackingCode)
    assert result[1].code in codes


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_events(client):
    result = client.get_tracking_code_events("BE058714266BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "BE058714266BR"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_events_without_city_field(client):
    result = client.get_tracking_code_events("FJ064849483BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "FJ064849483BR"
    assert result[0].events[0].city == ""


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_with_no_verification_digitevents(client):
    result = client.get_tracking_code_events("BE05871426BR")

    assert isinstance(result[0], TrackingCode)
    assert result[0].code == "BE058714266BR"


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_get_tracking_code_object_not_found_by_correios(client):
    tracking_code = client.get_tracking_code_events("DU05508759BR")[0]
    assert tracking_code.events

    event = tracking_code.events[0]
    assert isinstance(event, NotFoundTrackingEvent)
    assert event.timestamp
    assert event.status.type == "ERROR"
    assert event.status.status == 0


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_get_tracking_codes_events_over_limit(client):
    codes = ["DU05508759BR"] * 51
    with pytest.raises(TrackingCodesLimitExceededError):
        client.get_tracking_code_events(codes)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_builder_posting_card_status():
    builder = ModelBuilder()
    assert builder.build_posting_card_status("Normal") == PostingCard.ACTIVE
    assert builder.build_posting_card_status("Cancelado") == PostingCard.CANCELLED


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_posting_list_serialization(posting_list, shipping_label):
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert xml.startswith(b'<?xml version="1.0" encoding="ISO-8859-1"?><correioslog>')
    assert b"<codigo_servico_adicional>064</codigo_servico_adicional>" not in xml
    assert b"<valor_declarado>10,29</valor_declarado>" not in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_posting_list_serialization_with_crazy_utf8_character(posting_list, shipping_label):
    shipping_label.receiver.neighborhood = "Olho D’Água"
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert xml.startswith(b'<?xml version="1.0" encoding="ISO-8859-1"?><correioslog>')


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@pytest.mark.parametrize(
    "field, value, max_length",
    [
        ("complement", "Supermercado Manuel Rede Suprema", 30),
        ("street", "Travessa do Paralelepípedo de Santo Antonio da Cruz", 50),
    ],
)
def test_posting_list_serialization_max_length(field, value, max_length, posting_list, shipping_label):

    assert len(value) > max_length

    setattr(shipping_label.receiver, field, value)
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)

    assert value.encode("ISO-8859-1") not in xml
    assert value[:max_length].encode("ISO-8859-1") in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@pytest.mark.parametrize(
    "extra_service_vd,code",
    [(EXTRA_SERVICE_VD_PAC, b"064"), (EXTRA_SERVICE_VD_PAC_MINI, b"065"), (EXTRA_SERVICE_VD_SEDEX, b"019")],
)
def test_declared_value(extra_service_vd, code, posting_list, shipping_label):
    shipping_label.extra_services.append(ExtraService.get(extra_service_vd))
    shipping_label.real_value = 10.29
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert shipping_label.service == Service.get(SERVICE_PAC)
    assert b"<codigo_servico_adicional>%b</codigo_servico_adicional>" % code in xml
    assert b"<valor_declarado>20,50</valor_declarado>" in xml
    assert b"<valor_declarado>000,00</valor_declarado>" not in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_declared_value_without_insurance_extra_service(posting_list, shipping_label):
    shipping_label.real_value = 10.29
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)
    assert shipping_label.service == Service.get(SERVICE_PAC)
    assert b"<codigo_servico_adicional>19</codigo_servico_adicional>" not in xml
    assert b"<codigo_servico_adicional>64</codigo_servico_adicional>" not in xml
    assert b"<codigo_servico_adicional>65</codigo_servico_adicional>" not in xml
    assert b"<valor_declarado>000,00</valor_declarado>" in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_fail_empty_posting_list_serialization(posting_list):
    serializer = PostingListSerializer()
    with pytest.raises(PostingListSerializerError):
        serializer.get_document(posting_list)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_fail_closed_posting_list_serialization(posting_list, shipping_label):
    posting_list.add_shipping_label(shipping_label)
    posting_list.close_with_id(number=12345)

    serializer = PostingListSerializer()
    with pytest.raises(PostingListSerializerError):
        serializer.get_document(posting_list)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
def test_limit_size_city_name(posting_list, shipping_label):
    shipping_label.receiver.city = "Porto Alegre (Rio Grande do Sul)"
    shipping_label.sender.city = "Santa Maria (Rio Grande do Sul)"
    posting_list.add_shipping_label(shipping_label)
    serializer = PostingListSerializer()
    document = serializer.get_document(posting_list)
    serializer.validate(document)
    xml = serializer.get_xml(document)

    assert b"<cidade_destinatario><![CDATA[Porto Alegre (Rio Grande do Su]]></cidade_destinatario>" in xml
    assert b"<cidade_remetente><![CDATA[Santa Maria (Rio Grande do Sul]]></cidade_remetente>" in xml


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freights(client, posting_card, package):
    freights = client.calculate_freights(posting_card, [SERVICE_SEDEX, SERVICE_PAC], "07192100", "80030001", package)
    assert len(freights) == 2

    freight = freights[0]
    assert freight.error_code == 0
    assert not freight.error_message
    assert freight.service == SERVICE_SEDEX
    assert freight.delivery_time.days == 1
    assert freight.total == Decimal("23.75")
    assert freight.saturday is True
    assert freight.home is True

    freight = freights[1]
    assert freight.error_code == 0
    assert not freight.error_message
    assert freight.service == SERVICE_PAC
    assert freight.delivery_time.days == 6
    assert freight.total == Decimal("14.10")
    assert freight.saturday is False
    assert freight.home is True


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freights_with_extra_services(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip="07192100",
        to_zip="80030001",
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP],
    )
    assert len(freights) == 1

    freight = freights[0]
    assert freight.service == SERVICE_SEDEX
    assert freight.total == Decimal("96.03")
    assert freight.value == Decimal("23.75")
    assert freight.declared_value == Decimal("62.48")
    assert freight.mp_value == Decimal("5.50")
    assert freight.ar_value == Decimal("4.30")


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error(client, posting_card, package):
    package.real_weight = 80000  # invalid weight (80kg)
    freights = client.calculate_freights(posting_card, [SERVICE_SEDEX], "99999000", "99999999", package)
    assert len(freights) == 1
    assert freights[0].error_code == -4
    assert freights[0].error_message == "Peso excedido."


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_delivery_time(client):
    expected_delivery_time = 1
    delivery_time = client.calculate_delivery_time(Service.get(SERVICE_SEDEX), "07192100", "80030001")
    assert expected_delivery_time == int(delivery_time)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_delivery_time_service_not_allowed_for_path(client):
    expected_delivery_time = 0
    delivery_time = client.calculate_delivery_time(Service.get(SERVICE_PAC), "01311300", "01311300")
    assert expected_delivery_time == int(delivery_time)


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_10_restricted(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip="07192100",
        to_zip="09960610",
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP],
    )

    assert len(freights) == 1
    freight = freights[0]  # FreightResponse

    assert isinstance(freight, FreightResponse)
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 2
    assert freight.saturday


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_11_restricted(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip="09960610",
        to_zip="04475490",
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP],
    )

    assert len(freights) == 1

    freight = freights[0]
    assert isinstance(freight, FreightResponse)
    assert len(freights) == 1
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 8
    assert freight.saturday


@pytest.mark.skipif(not correios, reason="API Client support disabled")
@vcr.use_cassette
def test_calculate_freight_with_error_code_9_restricted(client, posting_card, package):
    freights = client.calculate_freights(
        posting_card=posting_card,
        services=[SERVICE_SEDEX],
        from_zip="09960610",
        to_zip="04475490",
        package=package,
        value="9000.00",
        extra_services=[EXTRA_SERVICE_AR, EXTRA_SERVICE_MP],
    )

    assert len(freights) == 1

    freight = freights[0]
    assert isinstance(freight, FreightResponse)
    assert freight.is_restricted_address()
    assert freight.error_code == FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED
    assert freight.value != 0
    assert freight.delivery_time.days == 8
    assert freight.saturday



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re

from vcr import VCR

USER_REGEX = re.compile(r"<usuario>\w+</usuario>")
PASS_REGEX = re.compile(r"<senha>.*</senha>")


def replace_auth(request):
    if not request.body:
        return request

    body = request.body.decode()
    body = USER_REGEX.sub(r"<usuario>teste</usuario>", body)
    body = PASS_REGEX.sub(r"<senha>****</senha>", body)
    request.body = body.encode()
    return request


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures")

vcr = VCR(
    record_mode="once",
    serializer="yaml",
    cassette_library_dir=os.path.join(FIXTURES_DIR, "cassettes"),
    path_transformer=VCR.ensure_suffix(".yaml"),
    match_on=["method"],
    before_record_request=replace_auth,
)
